#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as __get_version
from typing import Final

from .cli import main as register_plugin

main = register_plugin


def _get_version(name: str) -> str:
    """

    Parameters
    ----------
    name: str :


    Returns
    -------

    """
    try:
        return __get_version(name)
    except PackageNotFoundError:
        # Only occurs in development, since package is not installed properly
        return "0.0.0"


__version__: Final[str] = _get_version(__package__ or __name__)

__all__: list[str] = [main.__name__]



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


# Implementation of the PEP 440 version.
from dataclasses import dataclass, field
from functools import total_ordering
from typing import Annotated, Any, Final, Literal, Optional, cast, final

from annotated_types import Ge
from packaging.version import InvalidVersion
from packaging.version import Version as BaseVersion

from .logging import traced_function


@final
class VersionParserError(ValueError):
    """"""

    # Justification: Zen of python: Explicit is better than implicit
    pass  # pylint: disable=W0107


NonNegative = Ge(0)

NonNegativeInteger = Annotated[int, NonNegative]


@final
@dataclass(eq=False, order=False, frozen=True)
@total_ordering
class Version:
    """"""

    epoch: NonNegativeInteger = field(
        default=0, repr=True, hash=True, compare=True
    )
    release_tuple: tuple[NonNegativeInteger, ...] = field(default=(1, 0, 0))
    preview: Optional[
        tuple[
            Literal["a", "b", "c", "alpha", "beta", "rc"], NonNegativeInteger
        ]
    ] = field(default=None, repr=True, hash=True, compare=True)
    post: Optional[tuple[Literal["post"], NonNegativeInteger]] = field(
        default=None, repr=True, hash=True, compare=True
    )
    dev: Optional[tuple[Literal["dev"], NonNegativeInteger]] = field(
        default=None, repr=True, hash=True, compare=True
    )
    local: Optional[str] = field(
        default=None, repr=True, hash=True, compare=True
    )

    def __post_init__(self):
        if self.is_pre_release and not self.is_development_version:
            if self.preview[0] not in ["a", "b", "c", "alpha", "beta", "rc"]:
                raise ValueError(
                    f"Invalid pre-release identifier {self.preview[0]}"
                )

    @property
    def major(self) -> NonNegativeInteger:
        """"""
        return self.release_tuple[0] if len(self.release_tuple) >= 1 else 0

    @property
    def minor(self) -> NonNegativeInteger:
        """"""
        return self.release_tuple[1] if len(self.release_tuple) >= 2 else 0

    @property
    def micro(self) -> NonNegativeInteger:
        """"""
        return self.release_tuple[2] if len(self.release_tuple) >= 3 else 0

    @property
    def release(
        self,
    ) -> tuple[NonNegativeInteger, NonNegativeInteger, NonNegativeInteger]:
        """"""
        return (
            self.major,
            self.minor,
            self.micro,
        )

    @property
    def is_pre_release(self) -> bool:
        """"""
        return self.preview is not None or self.is_development_version

    @property
    def is_development_version(self) -> bool:
        """"""
        return self.dev is not None

    @property
    def is_post_release(self) -> bool:
        """"""
        return self.post is not None

    @property
    def is_local_version(self) -> bool:
        """"""
        return self.local is not None

    @property
    def is_alpha(self) -> bool:
        """"""
        alpha_part: Final[tuple[str, ...]] = ("a", "alpha")
        return self.preview is not None and self.__compare_preview(alpha_part)

    @property
    def is_beta(self) -> bool:
        """"""
        beta_part: Final[tuple[str, ...]] = ("b", "beta")
        return self.preview is not None and self.__compare_preview(beta_part)

    @property
    def is_release_candidate(self) -> bool:
        """"""
        rc_part: Final[tuple[str, ...]] = ("c", "rc")
        return self.preview is not None and self.__compare_preview(rc_part)

    @property
    def is_final(self) -> bool:
        """"""
        return (
            not self.preview
            and not self.is_local_version
            and not self.is_post_release
            and not self.is_development_version
        )

    def __compare_preview(self, valid_parts: tuple[str, ...]) -> bool:
        if self.preview is not None:
            pre: tuple[str, int] = cast(tuple[str, int], self.preview)
            return pre[0] in valid_parts
        return False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Version):
            raise ValueError(f"{other} must be an instance of Version")
        other_version: Version = cast(Version, other)
        return (
            self.epoch == other.epoch
            and self.release == other_version.release
            and self.preview == other_version.preview
            and self.post == other_version.post
            and self.dev == other_version.dev
            and self.local == other_version.local
        )

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Version):
            raise ValueError(f"{other} must be an instance of Version")
        other_version: Version = cast(Version, other)
        my_data = (
            self.epoch,
            self.release,
            self.preview,
            self.post,
            self.dev,
            self.local,
        )
        other_data = (
            other_version.epoch,
            other_version.release,
            other_version.preview,
            other_version.post,
            other_version.dev,
            other_version.local,
        )
        return my_data < other_data

    def __str__(self) -> str:
        return Pep440VersionFormatter().format(self)

    @staticmethod
    def default() -> "Version":
        """"""
        return Version(0, (1,), None, None, None, None)

    @staticmethod
    def from_string(version: str) -> "Version":
        """

        Parameters
        ----------
        version: str :


        Returns
        -------

        """
        try:
            _version: BaseVersion = BaseVersion(version)

            return Version(
                _version.epoch,
                _version.release,
                cast(
                    Optional[
                        tuple[
                            Literal["a", "b", "c", "alpha", "beta", "rc"], int
                        ]
                    ],
                    _version.pre,
                ),
                ("post", _version.post) if _version.post is not None else None,
                ("dev", _version.dev) if _version.dev is not None else None,
                _version.local,
            )
        except InvalidVersion as error:
            raise VersionParserError(
                f"{version} is not a valid version according to PEP 440."
            ) from error

    @staticmethod
    def can_parse_to_version(version: str) -> bool:
        """

        Parameters
        ----------
        version: str :


        Returns
        -------

        """
        try:
            _ = Version.from_string(version)
            return True
        except VersionParserError:
            return False


@final
# Justification: Only method to provide
class Pep440VersionFormatter:  # pylint: disable=R0903
    """"""

    @traced_function
    def format(self, version: Version) -> str:
        """

        Parameters
        ----------
        version: Version :


        Returns
        -------

        """
        parts: list[str] = []

        if version.epoch > 0:
            parts.append(f"{version.epoch}!")

        parts.append(".".join(str(part) for part in version.release))

        if version.preview is not None:
            parts.extend(str(part) for part in version.preview)

        if version.post is not None:
            parts.append(f".post{list(version.post)[1]}")

        if version.dev is not None:
            parts.append(f".dev{list(version.dev)[1]}")

        if version.is_local_version:
            parts.append(f"+{version.local}")

        return "".join(parts)



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from os import environ, pathsep
from pathlib import Path
from subprocess import run as stdlib_run_process
from sys import platform
from typing import Optional, Protocol, Union, cast

from ..core.logging import logger


class _CompletedProcessLike(Protocol):
    """"""

    @property
    def returncode(self) -> int:
        """"""
        raise NotImplementedError()

    @property
    def stdout(self) -> str:
        """"""
        raise NotImplementedError()

    @property
    def stderr(self) -> str:
        """"""
        raise NotImplementedError()


class _ProcessRunningCallable(Protocol):
    """"""

    def __call__(
        self,
        cmd: list[str],
        *,
        check: bool,
        capture_output: bool,
        cwd: Optional[Union[str, Path]],
        encoding: str = "utf-8",
    ) -> _CompletedProcessLike:
        raise NotImplementedError()


class ProcessRunner:
    """"""

    run_process: Optional[_ProcessRunningCallable] = None

    def _run_process(
        self,
        cmd: list[str],
        *,
        check: bool,
        capture_output: bool,
        cwd: Optional[Union[str, Path]],
        encoding: str = "utf-8",
    ) -> _CompletedProcessLike:
        """

        Parameters
        ----------
        cmd: list[str] :

        * :

        check: bool :

        capture_output: bool :

        cwd: Optional[Union[str, Path]] :

        encoding: str :
             (Default value = "utf-8")

        Returns
        -------

        """
        if self.run_process is not None:
            run_proc: _ProcessRunningCallable = cast(
                _ProcessRunningCallable, self.run_process
            )
            return run_proc(
                cmd,
                check=check,
                capture_output=capture_output,
                cwd=cwd,
                encoding=encoding,
            )

        return stdlib_run_process(
            cmd,
            check=check,
            capture_output=capture_output,
            cwd=cwd,
            encoding=encoding,
        )


class CliRunnerMixin(ProcessRunner):
    """"""

    def _which(
        self, exec_name: str, extension: Optional[str] = None
    ) -> Optional[Path]:
        """

        Parameters
        ----------
        exec_name: str :

        extension: Optional[str] :
             (Default value = None)

        Returns
        -------

        """
        search_path = environ["PATH"]
        logger.debug(
            "Searching for executable '%s' using search path '%s'",
            exec_name,
            search_path,
        )
        if search_path is None or len(search_path) == 0:
            return None

        extension = ".exe" if extension is None and platform == "win32" else ""
        executable_full_name = exec_name + extension
        paths = search_path.split(pathsep)
        for path in [Path(p) for p in paths]:
            logger.debug(
                "Searching for '%s' in '%s'", executable_full_name, path
            )
            file_path = path / executable_full_name
            if file_path.is_file():
                logger.debug("Found %s", file_path)
                return file_path

        logger.debug("Could not find %s", executable_full_name)
        return None

    def run(
        self,
        /,
        executable: Path,
        args: tuple[str, ...],
        *,
        raise_on_exit: bool = False,
        cwd: Optional[Path] = None,
    ) -> tuple[int, str, str]:
        """

        Parameters
        ----------
        / :

        executable: Path :

        args: tuple[str, ...] :

        * :

        raise_on_exit: bool :
             (Default value = False)
        cwd: Optional[Path] :
             (Default value = None)

        Returns
        -------

        """
        cmd = []
        cmd.append(str(executable))
        for arg in args:
            cmd.append(arg)

        logger.debug(
            "Running command '%s' with args [%s]",
            str(executable),
            ",".join(args),
        )

        completed: _CompletedProcessLike = self._run_process(
            cmd,
            check=raise_on_exit,
            capture_output=True,
            cwd=cwd,
            encoding="utf-8",
        )

        logger.debug("Process exited with code %d", completed.returncode)
        logger.debug(
            "Process wrote the following to stdout: \n%s", completed.stdout
        )
        logger.debug(
            "Process wrote the following to stderr: \n%s", completed.stderr
        )

        return (
            completed.returncode,
            completed.stdout,
            completed.stderr,
        )



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


import sys
from collections.abc import Iterable
from enum import Enum, IntEnum, auto
from functools import cached_property
from io import BytesIO
from pathlib import Path
from typing import Any, Final, Optional, Protocol, cast

from pyproject_metadata import StandardMetadata
from tomli_w import dump as dump_toml

from .logging import logger, traced_function

if sys.version_info >= (3, 10, 0):
    # suspicious mypy behavior
    from typing import TypeAlias  # type: ignore
else:
    from typing_extensions import TypeAlias

if sys.version_info >= (3, 11, 0):
    # suspicious mypy behavior
    from tomllib import load as load_toml  # type: ignore
else:
    # Python 3.11 -> mypy will not recognize this
    from tomli import load as load_toml  # type: ignore


_ConfigMapping: TypeAlias = dict[str, Any]


class _StringEnum(str, Enum):
    """"""

    # Justification: Zen of python: Explicit is better than implicit
    pass  # pylint: disable=W0107


# Justification: Minimal protocol
class ConfigHolder(Protocol):  # pylint: disable=R0903
    """"""

    root: Path
    PYPROJECT_FILENAME: str

    @property
    def config(self) -> _ConfigMapping:
        """"""
        # Method empty: Only a protocol stub
        raise NotImplementedError()


@traced_function
def _get_config_value(
    config: _ConfigMapping, *keys: str, default_value: Optional[Any] = None
) -> Optional[Any]:
    """

    Parameters
    ----------
    config: _ConfigMapping :

    *keys: str :

    default_value: Optional[Any] :
         (Default value = None)

    Returns
    -------

    """
    front: str

    key: str = ".".join(keys)
    logger.debug("Searching for '%s' in configuration", key)
    logger.debug("Configuration is set to: \n%s", config)

    while len(keys) > 1:
        front = keys[0]
        if front in config.keys():
            logger.debug("Found configuration section %s", front)
            config = cast(dict[str, Any], config[front])
            keys = tuple(keys[1:])
        else:
            logger.debug("Could not find configuration section %s.", front)
            return default_value

    front = keys[0]

    result = default_value if front not in config.keys() else config[front]
    logger.debug("Found value at '%s' is: %s", key, result)
    return result


@traced_function
def _set_config_value(config: _ConfigMapping, value: Any, *keys: str) -> None:
    """

    Parameters
    ----------
    config: _ConfigMapping :

    value: Any :

    *keys: str :


    Returns
    -------

    """
    front: str

    key: str = ".".join(keys)
    logger.debug("Setting '%s' to '%s'", key, value)

    while len(keys) > 1:
        front = keys[0]
        if front not in config.keys():
            logger.debug(
                "Key '%s' was not found. Adding empty configuration", front
            )
            config[front] = {}
        config = config[front]
        keys = tuple(keys[1:])

    front = keys[0]
    config[front] = value


class ConfigSections(_StringEnum):
    """"""

    PDM_BUMP: Final[str] = "pdm_bump"
    PDM_BUMP_VCS: Final[str] = "vcs"


class ConfigKeys(_StringEnum):
    """"""

    VERSION: Final[str] = "version"
    VERSION_SOURCE: Final[str] = "source"
    VERSION_SOURCE_FILE_PATH: Final[str] = "path"
    BUILD_BACKEND: Final[str] = "build-backend"
    VCS_PROVIDER: Final[str] = "provider"
    PROJECT_METADATA: Final[str] = "project"


class ConfigValues(_StringEnum):
    """"""

    VERSION_SOURCE_FILE: Final[str] = "file"
    VERSION_SOURCE_SCM: Final[str] = "scm"
    DEPREACTED_BUILD_BACKEND_PDM_PEP517_API: Final[str] = "pdm.pep517.api"
    BUILD_BACKEND_PDM_BACKEND: Final[str] = "pdm.backend"


# Justification: Currently no more meta data to check
class ProjectMetaData:  # pylint: disable=R0903
    """"""

    def __init__(self, meta_data: StandardMetadata) -> None:
        self.__meta_data = meta_data

    @property
    def is_dynamic_version(self) -> bool:
        """"""
        return ConfigKeys.VERSION in self.__meta_data.dynamic


class _ConfigSection(IntEnum):
    """"""

    ROOT = auto()
    METADATA = auto()
    PLUGIN_CONFIG = auto()
    BUILD_SYSTEM = auto()
    TOOL_CONFIG = auto()


class Config:
    """"""

    def __init__(self, project: ConfigHolder) -> None:
        self.__project: ConfigHolder = project

    @cached_property
    @traced_function
    def pyproject_file(self) -> Path:
        """"""
        return self.__project.root / self.__project.PYPROJECT_FILENAME

    @property
    @traced_function
    def meta_data(self) -> ProjectMetaData:
        """"""
        data: _ConfigMapping = self._get_pyproject_config(_ConfigSection.ROOT)
        meta: StandardMetadata = StandardMetadata.from_pyproject(data)
        return ProjectMetaData(meta)

    @traced_function
    def get_pyproject_metadata(self, *keys: tuple[str, ...]) -> Optional[Any]:
        """

        Parameters
        ----------
        *keys: tuple[str, ...] :


        Returns
        -------

        """
        config: _ConfigMapping = self._get_pyproject_config(
            _ConfigSection.METADATA
        )
        return _get_config_value(config, *keys)

    @traced_function
    def get_pyproject_build_system(
        self, *keys: tuple[str, ...]
    ) -> Optional[Any]:
        """

        Parameters
        ----------
        *keys: tuple[str, ...] :


        Returns
        -------

        """
        config: _ConfigMapping = self._get_pyproject_config(
            _ConfigSection.BUILD_SYSTEM
        )
        return _get_config_value(config, *keys)

    @traced_function
    def get_pyproject_tool_config(
        self, *keys: tuple[str, ...]
    ) -> Optional[Any]:
        """

        Parameters
        ----------
        *keys: tuple[str, ...] :


        Returns
        -------

        """
        config: _ConfigMapping = self._get_pyproject_config(
            _ConfigSection.TOOL_CONFIG
        )
        return _get_config_value(config, *keys)

    @traced_function
    def get_config_value(self, *keys: tuple[str, ...]) -> Optional[Any]:
        """

        Parameters
        ----------
        *keys: tuple[str, ...] :


        Returns
        -------

        """
        config: dict[str, Any] = self.__project.config
        return _get_config_value(config, *keys)

    @traced_function
    def get_config_or_pyproject_value(
        self, *keys: tuple[str, ...]
    ) -> Optional[Any]:
        """

        Parameters
        ----------
        *keys: tuple[str, ...] :


        Returns
        -------

        """
        config1: _ConfigMapping = self.__project.config
        config2: _ConfigMapping = self._get_pyproject_config(
            _ConfigSection.PLUGIN_CONFIG
        )

        return _get_config_value(config1, *keys) or _get_config_value(
            config2, *keys
        )

    @traced_function
    def set_pyproject_metadata(
        self, value: Any, *keys: tuple[str, ...]
    ) -> None:
        """

        Parameters
        ----------
        value: Any :

        *keys: tuple[str, ...] :


        Returns
        -------

        """
        config: _ConfigMapping = self._get_pyproject_config(
            _ConfigSection.ROOT
        )
        new_config: _ConfigMapping = _get_config_value(
            config, ConfigKeys.PROJECT_METADATA, default_value={}
        )
        _set_config_value(new_config, value, *keys)
        self._write_config(config)

    def _write_config(self, config: _ConfigMapping) -> None:
        """

        Parameters
        ----------
        config: _ConfigMapping :


        Returns
        -------

        """
        with BytesIO() as buffer:
            dump_toml(config, buffer)

            with open(self.pyproject_file, "wb+") as file_ptr:
                file_ptr.write(buffer.getvalue())
                logger.info(
                    "Successfully saved configuration to %s",
                    self.pyproject_file,
                )

    @traced_function
    def _get_pyproject_config(self, section: _ConfigSection) -> _ConfigMapping:
        """

        Parameters
        ----------
        section: _ConfigSection :


        Returns
        -------

        """
        project_data: _ConfigMapping = self._read_config()
        section_key: Iterable[str] = ()
        if section in (
            _ConfigSection.TOOL_CONFIG,
            _ConfigSection.PLUGIN_CONFIG,
        ):
            section_key = ["tool", "pdm"]
            if _ConfigSection.PLUGIN_CONFIG == section:
                section_key.extend(["bump-plugin"])
            section_key = tuple(section_key)
        elif _ConfigSection.BUILD_SYSTEM == section:
            section_key = ("build-system",)
        elif _ConfigSection.METADATA == section:
            section_key = ("project",)
        elif _ConfigSection.ROOT == section:
            return project_data or {}

        data = _get_config_value(project_data, *section_key, default_value={})

        return cast(_ConfigMapping, data)

    @traced_function
    def _read_config(self) -> dict[str, Any]:
        """"""
        project_file = self.pyproject_file
        with open(project_file, "rb") as file_pointer:
            return load_toml(file_pointer)



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#

from typing import Callable

from pdm_bump.actions import (
    VersionModifier,
)
from pdm_bump.actions.increment import (
    MajorIncrementingVersionModifier,
    MinorIncrementingVersionModifier,
    MicroIncrementingVersionModifier,
    FinalizingVersionModifier,
    EpochIncrementingVersionModifier,
    DevelopmentVersionIncrementingVersionModifier,
    PostVersionIncrementingVersionModifier,
)
from pdm_bump.actions.preview import (
    PreviewMismatchError,
    AlphaIncrementingVersionModifier,
    BetaIncrementingVersionModifier,
    ReleaseCandidateIncrementingVersionModifier,
)
from pdm_bump.actions.poetry_like import (
    PoetryLikePreReleaseVersionModifier,
    PoetryLikePreMajorVersionModifier,
    PoetryLikePreMinorVersionModifier,
    PoetryLikePrePatchVersionModifier,
)

from pdm_bump.core.version import Version

import pytest

parametrize = pytest.mark.parametrize
assert_raises = pytest.raises


class _UnitTestPersister:
    def save_version(self, version: Version) -> None:
        # Just for testing
        pass

_unit_test_persister = _UnitTestPersister()

_CREATE_NEXT_VERSION_PARAMS: list[
    tuple[str, str, str, Callable[[Version], Version]]
] = [
    (
        "Increment major parts with removing pre-parts and no pre-parts",
        "1.0.0",
        "2.0.0",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment major parts with removing pre-parts and alpha pre-parts",
        "1.0.0a1",
        "2.0.0",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment major parts with removing pre-parts and beta pre-parts",
        "1.0.0b1",
        "2.0.0",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment major parts with removing pre-parts and rc pre-parts",
        "1.0.0rc1",
        "2.0.0",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment major parts with removing pre-parts and dev parts",
        "1.0.0dev1",
        "2.0.0",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment major parts with removing pre-parts and post parts",
        "1.0.0post1",
        "2.0.0",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment major parts with removing pre-parts and local parts",
        "1.0.0+local",
        "2.0.0",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment major parts without removing pre-parts and no pre-parts",
        "1.0.0",
        "2.0.0",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment major parts without removing pre-parts and alpha pre-parts",
        "1.0.0a1",
        "2.0.0a1",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment major parts without removing pre-parts and beta pre-parts",
        "1.0.0b1",
        "2.0.0b1",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment major parts without removing pre-parts and rc pre-parts",
        "1.0.0rc1",
        "2.0.0rc1",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment major parts without removing pre-parts and dev parts",
        "1.0.0dev1",
        "2.0.0dev1",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment major parts without removing pre-parts and post parts",
        "1.0.0post1",
        "2.0.0post1",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment major parts without removing pre-parts and local parts",
        "1.0.0+local",
        "2.0.0+local",
        lambda v: MajorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment minor parts with removing pre-parts and no pre-parts",
        "1.0.0",
        "1.1.0",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment minor parts with removing pre-parts and alpha pre-parts",
        "1.0.0a1",
        "1.1.0",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment minor parts with removing pre-parts and beta pre-parts",
        "1.0.0b1",
        "1.1.0",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment minor parts with removing pre-parts and rc pre-parts",
        "1.0.0rc1",
        "1.1.0",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment minor parts with removing pre-parts and dev parts",
        "1.0.0dev1",
        "1.1.0",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment minor parts with removing pre-parts and post parts",
        "1.0.0post1",
        "1.1.0",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment minor parts with removing pre-parts and local parts",
        "1.0.0+local",
        "1.1.0",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment minor parts without removing pre-parts and no pre-parts",
        "1.0.0",
        "1.1.0",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment minor parts without removing pre-parts and alpha pre-parts",
        "1.0.0a1",
        "1.1.0a1",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment minor parVersionModifierts without removing pre-parts and beta pre-parts",
        "1.0.0b1",
        "1.1.0b1",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment minor parts without removing pre-parts and rc pre-parts",
        "1.0.0rc1",
        "1.1.0rc1",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment minor parts without removing pre-parts and dev parts",
        "1.0.0dev1",
        "1.1.0dev1",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment minor parts without removing pre-parts and post parts",
        "1.0.0post1",
        "1.1.0post1",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment minor parts without removing pre-parts and local parts",
        "1.0.0+local",
        "1.1.0+local",
        lambda v: MinorIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment micro parts with removing pre-parts and no pre-parts",
        "1.0.0",
        "1.0.1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment micro parts with removing pre-parts and alpha pre-parts",
        "1.0.0a1",
        "1.0.1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment micro parts with removing pre-parts and beta pre-parts",
        "1.0.0b1",
        "1.0.1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment micro parts with removing pre-parts and rc pre-parts",
        "1.0.0rc1",
        "1.0.1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment micro parts with removing pre-parts and dev parts",
        "1.0.0dev1",
        "1.0.1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment micro parts with removing pre-parts and post parts",
        "1.0.0post1",
        "1.0.1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment micro parts with removing pre-parts and local parts",
        "1.0.0+local",
        "1.0.1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment micro parts without removing pre-parts and no pre-parts",
        "1.0.0",
        "1.0.1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment micro parts without removing pre-parts and alpha pre-parts",
        "1.0.0a1",
        "1.0.1a1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment micro parts without removing pre-parts and beta pre-parts",
        "1.0.0b1",
        "1.0.1b1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment micro parts without removing pre-parts and rc pre-parts",
        "1.0.0rc1",
        "1.0.1rc1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment micro parts without removing pre-parts and dev parts",
        "1.0.0dev1",
        "1.0.1dev1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment micro parts without removing pre-parts and post parts",
        "1.0.0post1",
        "1.0.1post1",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment micro parts without removing pre-parts and local parts",
        "1.0.0+local",
        "1.0.1+local",
        lambda v: MicroIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment alpha parts without incrementing micro and providing no pre-release part",
        "1.0.0",
        "1.0.0a1",
        lambda v: AlphaIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment alpha parts without incrementing micro and providing a pre-release part",
        "1.0.0a1",
        "1.0.0a2",
        lambda v: AlphaIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment alpha parts with incrementing micro and providing no pre-release part",
        "1.0.0",
        "1.0.1a1",
        lambda v: AlphaIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment alpha parts with incrementing micro and providing a pre-release part",
        "1.0.0a1",
        "1.0.0a2",
        lambda v: AlphaIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment beta parts without incrementing micro and providing no pre-release part",
        "1.0.0",
        "1.0.0b1",
        lambda v: BetaIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment beta parts without incrementing micro and providing a pre-release part",
        "1.0.0b1",
        "1.0.0b2",
        lambda v: BetaIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment beta parts without incrementing micro and providing a pre-release part",
        "1.0.0a2",
        "1.0.0b1",
        lambda v: BetaIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment beta parts with incrementing micro and providing no pre-release part",
        "1.0.0",
        "1.0.1b1",
        lambda v: BetaIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment beta parts with incrementing micro and providing a pre-release part",
        "1.0.0b1",
        "1.0.0b2",
        lambda v: BetaIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment beta parts with incrementing micro and providing no pre-release part",
        "1.0.0a2",
        "1.0.0b1",
        lambda v: BetaIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment rc parts without incrementing micro and providing no pre-release part",
        "1.0.0",
        "1.0.0rc1",
        lambda v: ReleaseCandidateIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment rc parts without incrementing micro and providing a pre-release part",
        "1.0.0rc1",
        "1.0.0rc2",
        lambda v: ReleaseCandidateIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment rc parts without incrementing micro and providing no pre-release part",
        "1.0.0a2",
        "1.0.0rc1",
        lambda v: ReleaseCandidateIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment rc parts without incrementing micro and providing no pre-release part",
        "1.0.0b2",
        "1.0.0rc1",
        lambda v: ReleaseCandidateIncrementingVersionModifier(v, _unit_test_persister, False),
    ),
    (
        "Increment rc parts with incrementing micro and providing no pre-release part",
        "1.0.0",
        "1.0.1rc1",
        lambda v: ReleaseCandidateIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment rc parts with incrementing micro and providing a pre-release part",
        "1.0.0rc1",
        "1.0.0rc2",
        lambda v: ReleaseCandidateIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment rc parts with incrementing micro and providing no pre-release part",
        "1.0.0a2",
        "1.0.0rc1",
        lambda v: ReleaseCandidateIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment rc parts with incrementing micro and providing a pre-release part",
        "1.0.0b2",
        "1.0.0rc1",
        lambda v: ReleaseCandidateIncrementingVersionModifier(v, _unit_test_persister, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3",
        "1!1.2.3",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3a4",
        "1!1.2.3a4",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3b4",
        "1!1.2.3b4",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3rc4",
        "1!1.2.3rc4",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3dev5",
        "1!1.2.3dev5",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3post6",
        "1!1.2.3post6",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3+local7",
        "1!1.2.3+local7",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3b4post3dev6+local7",
        "1!1.2.3b4post3dev6+local7",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3",
        "1!1.2.3",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3a4",
        "1!1.2.3",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3b4",
        "1!1.2.3",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3rc4",
        "1!1.2.3",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3dev5",
        "1!1.2.3",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3post6",
        "1!1.2.3",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3+local7",
        "1!1.2.3",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3b4post3dev6+local7",
        "1!1.2.3",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, False),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3a4",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3b4",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3rc4",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3dev5",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3post6",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3+local7",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3b4post3dev6+local7",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, False, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3a4",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3b4",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3rc4",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3dev5",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3post6",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3+local7",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, True),
    ),
    (
        "Increment epoch without an epoch and without resetting version or removing any version part",
        "1.2.3b4post3dev6+local7",
        "1!1.0.0",
        lambda v: EpochIncrementingVersionModifier(v, _unit_test_persister, True, True),
    ),
    (
        "Remove non-final parts",
        "1.2.3",
        "1.2.3",
        lambda v: FinalizingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Remove non-final parts",
        "1.2.3a1",
        "1.2.3",
        lambda v: FinalizingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Remove non-final parts",
        "1.2.3b2",
        "1.2.3",
        lambda v: FinalizingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Remove non-final parts",
        "1.2.3rc3",
        "1.2.3",
        lambda v: FinalizingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Remove non-final parts",
        "1.2.3-dev1",
        "1.2.3",
        lambda v: FinalizingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Remove non-final parts",
        "1.2.3-post4",
        "1.2.3",
        lambda v: FinalizingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Remove non-final parts",
        "1.2.3+local8",
        "1.2.3",
        lambda v: FinalizingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Remove non-final parts",
        "1.2.3-b4-post6-dev8+local9",
        "1.2.3",
        lambda v: FinalizingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing development version",
        "1.2.3-dev1",
        "1.2.3-dev2",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing development version",
        "1.2.0a1",
        "1.2.0a2-dev1",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing development version",
        "1.2.0a1-dev2",
        "1.2.0a1-dev3",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing development version",
        "1.2.0b2",
        "1.2.0b3-dev1",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing development version",
        "1.2.0b2-dev1",
        "1.2.0b2-dev2",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing development version",
        "1.2.0rc1",
        "1.2.0rc2-dev1",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing development version",
        "1.2.0rc1-dev1",
        "1.2.0rc1-dev2",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of non-existing development version",
        "1.2.3",
        "1.2.4-dev1",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing post and non-existing development version",
        "1.2.3-post6",
        "1.2.4-dev1",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing post and existing development version",
        "1.2.3-post6-dev1",
        "1.2.3-post6-dev2",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing post and existing development version with local revision",
        "1.2.3-post6-dev1+local9",
        "1.2.3-post6-dev2+local9",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing development version with local revision",
        "1.2.3-dev1+local9",
        "1.2.3-dev2+local9",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment development part of existing post and non-existing development version with local revision",
        "1.2.3-post1+local9",
        "1.2.4-dev1+local9",
        lambda v: DevelopmentVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment post part of a non-existing post part",
        "1.2.3",
        "1.2.3-post1",
        lambda v: PostVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment post part of an existing post part",
        "1.2.3-post1",
        "1.2.3-post2",
        lambda v: PostVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment post part of a non-existing post part, but dev exists",
        "1.2.3-dev3",
        "1.2.3-post1",
        lambda v: PostVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment post part of an existing post part, but dev exists",
        "1.2.3-post1-dev3",
        "1.2.3-post2",
        lambda v: PostVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment post part of a non-existing post part, but local exists",
        "1.2.3+local1",
        "1.2.3-post1+local1",
        lambda v: PostVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Increment post part of a non-existing post part, but dev and local exist",
        "1.2.3-dev1+local1",
        "1.2.3-post1+local1",
        lambda v: PostVersionIncrementingVersionModifier(v, _unit_test_persister),
    ),
    (
        "Run premajor from poetry version commands (demo from https://python-poetry.org/docs/cli/#version)",
        "1.0.2",
        "2.0.0a0",
        lambda v: PoetryLikePreMajorVersionModifier(v, _unit_test_persister),
    ),
    (
        "Run preminor from poetry version commands (demo from https://python-poetry.org/docs/cli/#version)",
        "1.0.2",
        "1.1.0a0",
        lambda v: PoetryLikePreMinorVersionModifier(v, _unit_test_persister),
    ),
    (
        "Run prepatch from poetry version commands (demo from https://python-poetry.org/docs/cli/#version)",
        "1.0.2",
        "1.0.3a0",
        lambda v: PoetryLikePrePatchVersionModifier(v, _unit_test_persister),
    ),
    (
        "Run prerelease from poetry version commands (demo from https://python-poetry.org/docs/cli/#version)",
        "1.0.2",
        "1.0.3a0",
        lambda v: PoetryLikePreReleaseVersionModifier(v, _unit_test_persister),
    ),
    (
        "Run prerelease from poetry version commands (demo from https://python-poetry.org/docs/cli/#version)",
        "1.0.3a0",
        "1.0.3a1",
        lambda v: PoetryLikePreReleaseVersionModifier(v, _unit_test_persister),
    ),
    (
        "Run prerelease from poetry version commands (demo from https://python-poetry.org/docs/cli/#version)",
        "1.0.3b0",
        "1.0.3b1",
        lambda v: PoetryLikePreReleaseVersionModifier(v, _unit_test_persister),
    ),
    (
        "Run prerelease from poetry version commands (demo from https://python-poetry.org/docs/cli/#version)",
        "1.0.3rc0",
        "1.0.3rc1",
        lambda v: PoetryLikePreReleaseVersionModifier(v, _unit_test_persister),
    ),
]
_CREATE_NEXT_VERSION_ERROR_PARAMS: list[
    tuple[str, str, Callable[[Version], VersionModifier]]
] = [
    (
        "Increment alpha version if beta is present",
        "1.2.3b1",
        lambda v: AlphaIncrementingVersionModifier(v, _unit_test_persister, False),
        PreviewMismatchError,
    ),
    (
        "Increment alpha version if beta is present",
        "1.2.3b1",
        lambda v: AlphaIncrementingVersionModifier(v, _unit_test_persister, True),
        PreviewMismatchError,
    ),
    (
        "Increment alpha version if rc is present",
        "1.2.3rc1",
        lambda v: AlphaIncrementingVersionModifier(v, _unit_test_persister, False),
        PreviewMismatchError,
    ),
    (
        "Increment alpha version if rc is present",
        "1.2.3rc1",
        lambda v: AlphaIncrementingVersionModifier(v, _unit_test_persister, True),
        PreviewMismatchError,
    ),
    (
        "Increment beta version if rc is present",
        "1.2.3rc1",
        lambda v: BetaIncrementingVersionModifier(v, _unit_test_persister, False),
        PreviewMismatchError,
    ),
    (
        "Increment beta version if rc is present",
        "1.2.3rc1",
        lambda v: BetaIncrementingVersionModifier(v, _unit_test_persister, True),
        PreviewMismatchError,
    ),
    (
        "Pre-Major if version is dev version",
        "1.2.3-dev1",
        lambda v: PoetryLikePreMajorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Major if version is local version",
        "1.2.3+local17",
        lambda v: PoetryLikePreMajorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Major if version is post version",
        "1.2.3-post12",
        lambda v: PoetryLikePreMajorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Major if version is alpha version",
        "1.2.3a23",
        lambda v: PoetryLikePreMajorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Major if version is beta version",
        "1.2.3b23",
        lambda v: PoetryLikePreMajorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Major if version is release candidate",
        "1.2.3rc23",
        lambda v: PoetryLikePreMajorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Minor if version is dev version",
        "1.2.3-dev1",
        lambda v: PoetryLikePreMinorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Minor if version is local version",
        "1.2.3+local17",
        lambda v: PoetryLikePreMinorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Minor if version is post version",
        "1.2.3-post12",
        lambda v: PoetryLikePreMinorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Minor if version is alpha version",
        "1.2.3a23",
        lambda v: PoetryLikePreMinorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Minor if version is beta version",
        "1.2.3b23",
        lambda v: PoetryLikePreMinorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Minor if version is release candidate",
        "1.2.3rc23",
        lambda v: PoetryLikePreMinorVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Patch if version is dev version",
        "1.2.3-dev1",
        lambda v: PoetryLikePrePatchVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Patch if version is local version",
        "1.2.3+local17",
        lambda v: PoetryLikePrePatchVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Patch if version is post version",
        "1.2.3-post12",
        lambda v: PoetryLikePrePatchVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Patch if version is alpha version",
        "1.2.3a23",
        lambda v: PoetryLikePrePatchVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Patch if version is beta version",
        "1.2.3b23",
        lambda v: PoetryLikePrePatchVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
    (
        "Pre-Patch if version is release candidate",
        "1.2.3rc23",
        lambda v: PoetryLikePrePatchVersionModifier(v, _unit_test_persister),
        ValueError,
    ),
]

@parametrize(",".join(["message", "current_version_str", "expected_version_str", "factory"]), _CREATE_NEXT_VERSION_PARAMS)
def test_create_next_version_success(message, current_version_str, expected_version_str, factory) -> None:
    current: Version = Version.from_string(current_version_str)
    expected: Version = Version.from_string(expected_version_str)

    command: VersionModifier = factory(current)

    modified: Version = command.create_new_version()
    assert modified == expected

@parametrize(",".join(["message", "current_version_str", "factory", "exception_type"]), _CREATE_NEXT_VERSION_ERROR_PARAMS)
def test_create_next_version_fail(message, current_version_str, factory, exception_type) -> None:
    current: Version = Version.from_string(current_version_str)

    command: VersionModifier = factory(current)

    with assert_raises(exception_type):
        _ = command.create_new_version()



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""

# Justification: load module for decoupling
from . import explicit, increment, poetry_like, preview, vcs  # noqa: F401
from .base import ActionBase, VersionConsumer, VersionModifier, actions

__all__ = [
    "actions",
    "ActionBase",
    "VersionConsumer",
    "VersionModifier",
]



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#

""""""


# Justification: load module for decoupling
from . import gitcli  # noqa: F401
from .core import (
    DefaultVcsProvider,
    VcsProvider,
    VcsProviderAggregator,
    VcsProviderFactory,
    VcsProviderRegistry,
    vcs_provider,
    vcs_providers,
)
from .history import (
    Commit,
    CommitStatistics,
    CommitType,
    ConventionalCommitParser,
    History,
)

__all__ = (
    "vcs_provider",
    "vcs_providers",
    "VcsProvider",
    "VcsProviderAggregator",
    "VcsProviderFactory",
    "VcsProviderRegistry",
    "DefaultVcsProvider",
    "Commit",
    "CommitType",
    "CommitStatistics",
    "History",
    "ConventionalCommitParser",
)



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#

from typing import Final, final

from pdm_bump.core.version import (
    Version,
    VersionParserError,
    Pep440VersionFormatter,
    BaseVersion,
)

import pytest

parameterize = pytest.mark.parametrize
assert_raises = pytest.raises


_SUB_TESTS_VERSION_FROM_STR_RLS: Final[
    list[tuple[str, str, tuple[int, int, int]]]
] = [
    ("Regular default version", "1.0.0", (1, 0, 0)),
    ("Regular default version no micro", "1.0", (1, 0, 0)),
    ("Regular default version no minor and no micro", "1", (1, 0, 0)),
    ("Pre-fixed default version", "v1.0.0", (1, 0, 0)),
    ("Pre-fixed default version no micro", "v1.0", (1, 0, 0)),
    ("Pre-fixed default version no minor and no micro", "v1", (1, 0, 0)),
    ("Regular non-final version", "0.1.0", (0, 1, 0)),
    ("Regular non-final version no micro", "0.1.0", (0, 1, 0)),
    ("Pre-fixed non-final version", "v0.1.0", (0, 1, 0)),
    ("Pre-fixed non-final version no micro", "v0.1", (0, 1, 0)),
    ("Regular version", "6.2.13", (6, 2, 13)),
    ("Regular version no micro", "6.2", (6, 2, 0)),
    ("Regular version no minor and no micro", "6", (6, 0, 0)),
    ("Pre-fixed version", "v6.2.13", (6, 2, 13)),
    ("Pre-fixed version no micro", "v6.2", (6, 2, 0)),
    ("Pre-fixed version no minor and no micro", "v6", (6, 0, 0)),
    ("Regular double-digit version", "19.22.83", (19, 22, 83)),
    ("Regular double-digit version no micro", "19.22", (19, 22, 0)),
    (
        "Regular double-digit version no minor and no micro",
        "19",
        (19, 0, 0),
    ),
    ("Pre-fixed double-digit version", "v19.22.83", (19, 22, 83)),
    ("Pre-fixed double-digit version no micro", "v19.22", (19, 22, 0)),
    (
        "Pre-fixed double-digit version no minor and no micro",
        "v19",
        (19, 0, 0),
    ),
]
_SUB_TESTS_VERSION_FROM_STR_EPOCH: Final[list[tuple[str, str, int]]] = [
    ("No epoch set", "1.0.0", 0),
    ("Epoch is one", "1!1.0.0", 1),
    ("Epoch is double-digit", "17!1.0.0", 17),
]
_SUB_TESTS_VERSION_FROM_STR_PRE: Final[
    list[tuple[str, str, tuple[str, int]]]
] = [
    ("None: Alpha short single digit no prefix", "0.1.0a1", ("a", 1)),
    ("None: Alpha long single digit no prefix", "0.1.0alpha1", ("a", 1)),
    ("None: Alpha short single digit prefix", "v0.1.0a1", ("a", 1)),
    ("None: Alpha long single digit prefix", "v0.1.0alpha1", ("a", 1)),
    ("None: Alpha short double digit no prefix", "0.1.0a23", ("a", 23)),
    ("None: Alpha long double digit no prefix", "0.1.0alpha23", ("a", 23)),
    ("None: Alpha short double digit prefix", "v0.1.0a23", ("a", 23)),
    ("None: Alpha long double digit prefix", "v0.1.0alpha23", ("a", 23)),
    ("None: Beta short single digit no prefix", "0.1.0b1", ("b", 1)),
    ("None: Beta long single digit no prefix", "0.1.0beta1", ("b", 1)),
    ("None: Beta short single digit prefix", "v0.1.0b1", ("b", 1)),
    ("None: Beta long single digit prefix", "v0.1.0beta1", ("b", 1)),
    ("None: Beta short double digit no prefix", "0.1.0b42", ("b", 42)),
    ("None: Beta long double digit no prefix", "0.1.0beta42", ("b", 42)),
    ("None: Beta short double digit prefix", "v0.1.0b42", ("b", 42)),
    ("None: Beta long double digit prefix", "v0.1.0beta42", ("b", 42)),
    (
        "None: Release-candidate short single digit no prefix",
        "0.1.0rc1",
        ("rc", 1),
    ),
    (
        "None: Release-candidate long single digit no prefix",
        "0.1.0rc1",
        ("rc", 1),
    ),
    (
        "None: Release-candidate short single digit prefix",
        "v0.1.0c1",
        ("rc", 1),
    ),
    (
        "None: Release-candidate long single digit prefix",
        "v0.1.0rc1",
        ("rc", 1),
    ),
    (
        "None: Release-candidate short double digit no prefix",
        "0.1.0rc99",
        ("rc", 99),
    ),
    (
        "None: Release-candidate long double digit no prefix",
        "0.1.0rc99",
        ("rc", 99),
    ),
    (
        "None: Release-candidate short double digit prefix",
        "v0.1.0c99",
        ("rc", 99),
    ),
    (
        "None: Release-candidate long double digit prefix",
        "v0.1.0rc99",
        ("rc", 99),
    ),
    ("Dot: Alpha short single digit no prefix", "0.1.0.a1", ("a", 1)),
    ("Dot: Alpha long single digit no prefix", "0.1.0.alpha1", ("a", 1)),
    ("Dot: Alpha short single digit prefix", "v0.1.0.a1", ("a", 1)),
    ("Dot: Alpha long single digit prefix", "v0.1.0.alpha1", ("a", 1)),
    ("Dot: Alpha short double digit no prefix", "0.1.0.a23", ("a", 23)),
    ("Dot: Alpha long double digit no prefix", "0.1.0.alpha23", ("a", 23)),
    ("Dot: Alpha short double digit prefix", "v0.1.0.a23", ("a", 23)),
    ("Dot: Alpha long double digit prefix", "v0.1.0.alpha23", ("a", 23)),
    ("Dot: Beta short single digit no prefix", "0.1.0.b1", ("b", 1)),
    ("Dot: Beta long single digit no prefix", "0.1.0.beta1", ("b", 1)),
    ("Dot: Beta short single digit prefix", "v0.1.0.b1", ("b", 1)),
    ("Dot: Beta long single digit prefix", "v0.1.0.beta1", ("b", 1)),
    ("Dot: Beta short double digit no prefix", "0.1.0.b42", ("b", 42)),
    ("Dot: Beta long double digit no prefix", "0.1.0.beta42", ("b", 42)),
    ("Dot: Beta short double digit prefix", "v0.1.0.b42", ("b", 42)),
    ("Dot: Beta long double digit prefix", "v0.1.0.beta42", ("b", 42)),
    (
        "Dot: Release-candidate short single digit no prefix",
        "0.1.0.rc1",
        ("rc", 1),
    ),
    (
        "Dot: Release-candidate long single digit no prefix",
        "0.1.0.rc1",
        ("rc", 1),
    ),
    (
        "Dot: Release-candidate short single digit prefix",
        "v0.1.0.c1",
        ("rc", 1),
    ),
    (
        "Dot: Release-candidate long single digit prefix",
        "v0.1.0.rc1",
        ("rc", 1),
    ),
    (
        "Dot: Release-candidate short double digit no prefix",
        "0.1.0.rc99",
        ("rc", 99),
    ),
    (
        "Dot: Release-candidate long double digit no prefix",
        "0.1.0.rc99",
        ("rc", 99),
    ),
    (
        "Dot: Release-candidate short double digit prefix",
        "v0.1.0.c99",
        ("rc", 99),
    ),
    (
        "Dot: Release-candidate long double digit prefix",
        "v0.1.0.rc99",
        ("rc", 99),
    ),
    ("Hyphen: Alpha short single digit no prefix", "0.1.0-a1", ("a", 1)),
    (
        "Hyphen: Alpha long single digit no prefix",
        "0.1.0-alpha1",
        ("a", 1),
    ),
    ("Hyphen: Alpha short single digit prefix", "v0.1.0-a1", ("a", 1)),
    ("Hyphen: Alpha long single digit prefix", "v0.1.0-alpha1", ("a", 1)),
    ("Hyphen: Alpha short double digit no prefix", "0.1.0-a23", ("a", 23)),
    (
        "Hyphen: Alpha long double digit no prefix",
        "0.1.0-alpha23",
        ("a", 23),
    ),
    ("Hyphen: Alpha short double digit prefix", "v0.1.0-a23", ("a", 23)),
    (
        "Hyphen: Alpha long double digit prefix",
        "v0.1.0-alpha23",
        ("a", 23),
    ),
    ("Hyphen: Beta short single digit no prefix", "0.1.0-b1", ("b", 1)),
    ("Hyphen: Beta long single digit no prefix", "0.1.0-beta1", ("b", 1)),
    ("Hyphen: Beta short single digit prefix", "v0.1.0-b1", ("b", 1)),
    ("Hyphen: Beta long single digit prefix", "v0.1.0-beta1", ("b", 1)),
    ("Hyphen: Beta short double digit no prefix", "0.1.0-b42", ("b", 42)),
    (
        "Hyphen: Beta long double digit no prefix",
        "0.1.0-beta42",
        ("b", 42),
    ),
    ("Hyphen: Beta short double digit prefix", "v0.1.0-b42", ("b", 42)),
    ("Hyphen: Beta long double digit prefix", "v0.1.0-beta42", ("b", 42)),
    (
        "Hyphen: Release-candidate short single digit no prefix",
        "0.1.0-rc1",
        ("rc", 1),
    ),
    (
        "Hyphen: Release-candidate long single digit no prefix",
        "0.1.0-rc1",
        ("rc", 1),
    ),
    (
        "Hyphen: Release-candidate short single digit prefix",
        "v0.1.0-c1",
        ("rc", 1),
    ),
    (
        "Hyphen: Release-candidate long single digit prefix",
        "v0.1.0-rc1",
        ("rc", 1),
    ),
    (
        "Hyphen: Release-candidate short double digit no prefix",
        "0.1.0-rc99",
        ("rc", 99),
    ),
    (
        "Hyphen: Release-candidate long double digit no prefix",
        "0.1.0-rc99",
        ("rc", 99),
    ),
    (
        "Hyphen: Release-candidate short double digit prefix",
        "v0.1.0-c99",
        ("rc", 99),
    ),
    (
        "Hyphen: Release-candidate long double digit prefix",
        "v0.1.0-rc99",
        ("rc", 99),
    ),
    (
        "Underscore: Alpha short single digit no prefix",
        "0.1.0_a1",
        ("a", 1),
    ),
    (
        "Underscore: Alpha long single digit no prefix",
        "0.1.0_alpha1",
        ("a", 1),
    ),
    ("Underscore: Alpha short single digit prefix", "v0.1.0_a1", ("a", 1)),
    (
        "Underscore: Alpha long single digit prefix",
        "v0.1.0_alpha1",
        ("a", 1),
    ),
    (
        "Underscore: Alpha short double digit no prefix",
        "0.1.0_a23",
        ("a", 23),
    ),
    (
        "Underscore: Alpha long double digit no prefix",
        "0.1.0_alpha23",
        ("a", 23),
    ),
    (
        "Underscore: Alpha short double digit prefix",
        "v0.1.0_a23",
        ("a", 23),
    ),
    (
        "Underscore: Alpha long double digit prefix",
        "v0.1.0_alpha23",
        ("a", 23),
    ),
    (
        "Underscore: Beta short single digit no prefix",
        "0.1.0_b1",
        ("b", 1),
    ),
    (
        "Underscore: Beta long single digit no prefix",
        "0.1.0_beta1",
        ("b", 1),
    ),
    ("Underscore: Beta short single digit prefix", "v0.1.0_b1", ("b", 1)),
    (
        "Underscore: Beta long single digit prefix",
        "v0.1.0_beta1",
        ("b", 1),
    ),
    (
        "Underscore: Beta short double digit no prefix",
        "0.1.0_b42",
        ("b", 42),
    ),
    (
        "Underscore: Beta long double digit no prefix",
        "0.1.0_beta42",
        ("b", 42),
    ),
    (
        "Underscore: Beta short double digit prefix",
        "v0.1.0_b42",
        ("b", 42),
    ),
    (
        "Underscore: Beta long double digit prefix",
        "v0.1.0_beta42",
        ("b", 42),
    ),
    (
        "Underscore: Release-candidate short single digit no prefix",
        "0.1.0_rc1",
        ("rc", 1),
    ),
    (
        "Underscore: Release-candidate long single digit no prefix",
        "0.1.0_rc1",
        ("rc", 1),
    ),
    (
        "Underscore: Release-candidate short single digit prefix",
        "v0.1.0_c1",
        ("rc", 1),
    ),
    (
        "Underscore: Release-candidate long single digit prefix",
        "v0.1.0_rc1",
        ("rc", 1),
    ),
    (
        "Underscore: Release-candidate short double digit no prefix",
        "0.1.0_rc99",
        ("rc", 99),
    ),
    (
        "Underscore: Release-candidate long double digit no prefix",
        "0.1.0_rc99",
        ("rc", 99),
    ),
    (
        "Underscore: Release-candidate short double digit prefix",
        "v0.1.0_c99",
        ("rc", 99),
    ),
    (
        "Underscore: Release-candidate long double digit prefix",
        "v0.1.0_rc99",
        ("rc", 99),
    ),
]
_SUB_TESTS_VERSION_FROM_STR_DEV: Final[
    list[tuple[str, str, tuple[str, int]]]
] = [
    ("Regular version zero digit none", "0.0.1dev0", ("dev", 0)),
    ("Regular version zero digit dot", "0.0.1.dev0", ("dev", 0)),
    ("Regular version zero digit hyphen", "0.0.1-dev0", ("dev", 0)),
    ("Regular version zero digit underscore", "0.0.1_dev0", ("dev", 0)),
    ("Regular version no digit none", "0.0.1dev", ("dev", 0)),
    ("Regular version no digit dot", "0.0.1.dev", ("dev", 0)),
    ("Regular version no digit hyphen", "0.0.1-dev", ("dev", 0)),
    ("Regular version no digit underscore", "0.0.1_dev", ("dev", 0)),
    ("Regular version single digit none", "0.0.1dev1", ("dev", 1)),
    ("Regular version single digit dot", "0.0.1.dev1", ("dev", 1)),
    ("Regular version single digit hyphen", "0.0.1-dev1", ("dev", 1)),
    ("Regular version single digit underscore", "0.0.1_dev1", ("dev", 1)),
    ("Regular version double digit none", "0.0.1dev57", ("dev", 57)),
    ("Regular version double digit dot", "0.0.1.dev57", ("dev", 57)),
    ("Regular version double digit hyphen", "0.0.1-dev57", ("dev", 57)),
    (
        "Regular version double digit underscore",
        "0.0.1_dev57",
        ("dev", 57),
    ),
    ("Pre-fixed version single digit none", "v0.0.1dev1", ("dev", 1)),
    ("Pre-fixed version single digit dot", "v0.0.1.dev1", ("dev", 1)),
    ("Pre-fixed version single digit hyphen", "v0.0.1-dev1", ("dev", 1)),
    (
        "Pre-fixed version single digit underscore",
        "v0.0.1_dev1",
        ("dev", 1),
    ),
    ("Pre-fixed version double digit none", "v0.0.1dev57", ("dev", 57)),
    ("Pre-fixed version double digit dot", "v0.0.1.dev57", ("dev", 57)),
    ("Pre-fixed version double digit hyphen", "v0.0.1-dev57", ("dev", 57)),
    (
        "Pre-fixed version double digit underscore",
        "v0.0.1_dev57",
        ("dev", 57),
    ),
]
_SUB_TESTS_VERSION_FROM_STR_POST: Final[
    list[tuple[str, str, tuple[str, int]]]
] = [
    ("Regular version zero digit none", "2.0.0post0", ("post", 0)),
    ("Regular version zero digit dot", "2.0.0.post0", ("post", 0)),
    ("Regular version zero digit hyphen", "2.0.0-post0", ("post", 0)),
    ("Regular version zero digit underscore", "2.0.0_post0", ("post", 0)),
    ("Regular version no digit none", "2.0.0post", ("post", 0)),
    ("Regular version no digit dot", "2.0.0.post", ("post", 0)),
    ("Regular version no digit hyphen", "2.0.0-post", ("post", 0)),
    ("Regular version no digit underscore", "2.0.0_post", ("post", 0)),
    ("Regular version single digit none", "2.0.0post7", ("post", 7)),
    ("Regular version single digit dot", "2.0.0.post7", ("post", 7)),
    ("Regular version single digit hyphen", "2.0.0-post7", ("post", 7)),
    (
        "Regular version single digit underscore",
        "2.0.0_post7",
        ("post", 7),
    ),
    ("Regular version double digit none", "2.0.0post63", ("post", 63)),
    ("Regular version double digit dot", "2.0.0.post63", ("post", 63)),
    ("Regular version double digit hyphen", "2.0.0-post63", ("post", 63)),
    (
        "Regular version double digit underscore",
        "2.0.0_post63",
        ("post", 63),
    ),
    ("Regular version single digit special form", "2.0.0-3", ("post", 3)),
    (
        "Regular version double digit special form",
        "2.0.0-78",
        ("post", 78),
    ),
    ("Pre-fixed version zero digit none", "v2.0.0post0", ("post", 0)),
    ("Pre-fixed version zero digit dot", "v2.0.0.post0", ("post", 0)),
    ("Pre-fixed version zero digit hyphen", "v2.0.0-post0", ("post", 0)),
    (
        "Pre-fixed version zero digit underscore",
        "v2.0.0_post0",
        ("post", 0),
    ),
    ("Pre-fixed version no digit none", "v2.0.0post", ("post", 0)),
    ("Pre-fixed version no digit dot", "v2.0.0.post", ("post", 0)),
    ("Pre-fixed version no digit hyphen", "v2.0.0-post", ("post", 0)),
    ("Pre-fixed version no digit underscore", "v2.0.0_post", ("post", 0)),
    ("Pre-fixed version single digit none", "v2.0.0post7", ("post", 7)),
    ("Pre-fixed version single digit dot", "v2.0.0.post7", ("post", 7)),
    ("Pre-fixed version single digit hyphen", "v2.0.0-post7", ("post", 7)),
    (
        "Pre-fixed version single digit underscore",
        "v2.0.0_post7",
        ("post", 7),
    ),
    ("Pre-fixed version double digit none", "v2.0.0post63", ("post", 63)),
    ("Pre-fixed version double digit dot", "v2.0.0.post63", ("post", 63)),
    (
        "Pre-fixed version double digit hyphen",
        "v2.0.0-post63",
        ("post", 63),
    ),
    (
        "Pre-fixed version double digit underscore",
        "v2.0.0_post63",
        ("post", 63),
    ),
    (
        "Pre-fixed version single digit special form",
        "v2.0.0-3",
        ("post", 3),
    ),
    (
        "Pre-fixed version double digit special form",
        "v2.0.0-78",
        ("post", 78),
    ),
]  # TODO: rev/r
_SUB_TESTS_VERSION_FROM_STR_LOCAL: Final[list[tuple[str, str, str]]] = [
    ("Regular version single digit plus", "3.1.4+2", "2"),
    ("Regular version double digit plus", "3.1.4+12", "12"),
    ("Regular version alpha-numeric plus", "3.1.4+af2b", "af2b"),
    (
        "Regular version hyphen plus",
        "3.1.4+just-my-2-cents",
        "just.my.2.cents",
    ),
    (
        "Regular version hyphen underscore plus",
        "3.1.4+just_my-2_cents",
        "just.my.2.cents",
    ),
    ("Pre-fixed version single digit plus", "v3.1.4+2", "2"),
    ("Pre-fixed version double digit plus", "v3.1.4+12", "12"),
    ("Pre-fixed version alpha-numeric plus", "v3.1.4+af2b", "af2b"),
    (
        "Pre-fixed version hyphen plus",
        "v3.1.4+just-my-2-cents",
        "just.my.2.cents",
    ),
    (
        "Pre-fixed version hyphen underscore plus",
        "v3.1.4+just_my-2_cents",
        "just.my.2.cents",
    ),
]
_SUB_TESTS_VERSION_PARSE_SUCCESS: Final[list[tuple[str, str]]] = [
    ("Regular version", "1.2.3"),
    ("Prefixed version", "v1.2.3"),
    ("Prefixed version with alpha", "v1.2.3a4"),
    ("Prefixed version with beta", "v1.2.3b4"),
    ("Prefixed version with rc", "v1.2.3rc4"),
    ("Prefixed version with rc as post", "v1.2.3a4post5"),
    ("Prefixed version with rc as post alternative", "v1.2.3a4-5"),
    ("Prefixed version with rc as post and dev", "v1.2.3a4-post5.dev17"),
    (
        "Prefixed version with rc as post and dev and local",
        "v1.2.3a4-post5.dev17+erg.13",
    ),
]
_SUB_TESTS_VERSION_PARSE_FAIL: Final[list[tuple[str, str]]] = [
    ("Wrong prefix", "a0.1.0+2"),
    ("Negative epoch", "-2!v0.1.0+2"),
    ("Alpha epoch", "f!v0.1.0+2"),
    ("Wrong separators", "0-1-0"),
    ("Wrong alpha", "0.1.0alp1"),
    ("Wrong beta", "0.1.0bet2"),
    ("Wrong release candidate", "0.1.0relca3"),
    ("Negative major", "-1.0.0"),
    ("Negative minor", "1.-2.0"),
    ("Negative micro", "1.2.-3"),
]

def test_default_version_is_one() -> None:
    v1: Version = Version.default()
    v2: Version = Version.from_string("1")
    assert v1 == v2, "Versions match"

@parameterize(",".join(["message", "version", "epoch"]), _SUB_TESTS_VERSION_FROM_STR_EPOCH)
def test_create_version_from_str_epoch_part(message, version, epoch) -> None:
    v: Version = Version.from_string(version)
    assert v.epoch == epoch

@parameterize(",".join(["message", "version", "release"]), _SUB_TESTS_VERSION_FROM_STR_RLS)
def test_create_version_from_str_rls_part(message, version, release) -> None:
    v: Version = Version.from_string(version)
    assert v.release == release

@parameterize(",".join(["message", "version", "preview"]), _SUB_TESTS_VERSION_FROM_STR_PRE)
def test_create_version_from_str_preview_part(message, version, preview) -> None:
    v: Version = Version.from_string(version)
    assert v.preview == preview

@parameterize(",".join(["message", "version", "dev"]), _SUB_TESTS_VERSION_FROM_STR_DEV)
def test_create_version_from_str_dev_part(message, version, dev) -> None:
    v: Version = Version.from_string(version)
    assert v.dev == dev

@parameterize(",".join(["message", "version", "post"]), _SUB_TESTS_VERSION_FROM_STR_POST)
def test_create_version_from_str_post_part(message, version, post) -> None:
    v: Version = Version.from_string(version)
    assert v.post == post

@parameterize(",".join(["message", "version", "local"]), _SUB_TESTS_VERSION_FROM_STR_LOCAL)
def test_create_version_from_str_local_part(message, version, local) -> None:
    v: Version = Version.from_string(version)
    assert v.local == local

@parameterize(",".join(["message", "version"]), _SUB_TESTS_VERSION_PARSE_SUCCESS)
def test_create_version_from_str_can_parse_success(message, version) -> None:
    result: bool = Version.can_parse_to_version(version)
    assert result is True

@parameterize(",".join(["message", "version"]), _SUB_TESTS_VERSION_PARSE_FAIL)
def test_create_version_from_str_can_parse_fail(message, version) -> None:
    result: bool = Version.can_parse_to_version(version)
    assert result is False

@parameterize(",".join(["message", "version"]), _SUB_TESTS_VERSION_PARSE_FAIL)
def test_create_version_from_str_parse_raises(message, version) -> None:
    with assert_raises(VersionParserError):
        Version.from_string(version)

def test_version_major() -> None:
    v = Version(release_tuple=(1, 2, 3))
    assert v.major == 1

def test_version_minor() -> None:
    v = Version(release_tuple=(1, 2, 3))
    assert v.minor == 2

def test_version_micro() -> None:
    v = Version(release_tuple=(1, 2, 3))
    assert v.micro == 3

def test_version_major_unset() -> None:
    v = Version(release_tuple=tuple())
    assert v.major == 0

def test_version_minor_unset() -> None:
    v = Version(release_tuple=(1,))
    assert v.minor == 0

def test_version_micro_unset() -> None:
    v = Version(release_tuple=(1, 2))
    assert v.micro == 0

def test_version_is_no_pre_release() -> None:
    v = Version(release_tuple=(1, 2, 3))
    _assure_pre_matches(v, False, False, False, False, False)
    assert (v.is_post_release or v.is_local_version) is False

def test_version_is_alpha_true() -> None:
    v = _create_version(as_alpha=True)
    _assure_pre_matches(v, True, False, False, True, False)
    assert (v.is_post_release or v.is_local_version) is False

def test_version_is_beta_true() -> None:
    v = _create_version(as_beta=True)
    _assure_pre_matches(v, False, True, False, True, False)
    assert (v.is_post_release or v.is_local_version) is False

def test_version_is_rc_true() -> None:
    v = _create_version(as_rc=True)
    _assure_pre_matches(v, False, False, True, True, False)
    assert (v.is_post_release or v.is_local_version) is False

def test_version_is_dev_true() -> None:
    v = _create_version(as_dev=True)
    _assure_pre_matches(v, False, False, False, True, True)
    assert (v.is_post_release or v.is_local_version) is False

def test_version_is_post_true() -> None:
    v = _create_version(as_post=True)
    _assure_pre_matches(v, False, False, False, False, False)
    assert v.is_post_release is True
    assert (v.is_post_release and v.is_local_version) is False

def test_version_is_local_true() -> None:
    v = _create_version(as_local=True)
    _assure_pre_matches(v, False, False, False, False, False)
    assert v.is_local_version is True
    assert (v.is_post_release and v.is_local_version) is False

def _create_version(
    *,
    as_alpha: bool = False,
    as_beta: bool = False,
    as_rc: bool = False,
    as_dev: bool = False,
    as_post: bool = False,
    as_local: bool = False,
) -> Version:
    epoch = 1
    release = (2, 3, 4)
    dev = None if not as_dev else ("dev", 5)
    post = None if not as_post else ("post", 6)
    local = None if not as_local else "as.local7"

    items = int(as_alpha) + int(as_beta) + int(as_rc)
    if items > 1:
        raise ValueError("Only one pre-release identifier can be set")
    pre = None
    if as_alpha:
        pre = ("a", 8)
    elif as_beta:
        pre = ("b", 9)
    elif as_rc:
        pre = ("rc", 10)

    return Version(epoch, release, pre, post, dev, local)

def _assure_pre_matches(
    version: Version, is_alpha, is_beta, is_rc, is_pre, is_dev
) -> None:
    assert version.is_alpha == is_alpha, "Alpha part matches"
    assert version.is_beta == is_beta, "Beta part matches"
    assert version.is_release_candidate == is_rc, "RC part matches"
    assert version.is_pre_release == is_pre, "pre-release matches"
    assert version.is_development_version == is_dev, "Development matches"


import unittest


@final
class CreateVersionTest(unittest.TestCase):
    pass  # TODO implement


@final
class VersionOrderingTest(unittest.TestCase):
    pass  # TODO implement


@final
class FormatVersionTest(unittest.TestCase):
    pass  # TODO implement


@final
class RoundtripTest(unittest.TestCase):
    pass  # TODO implement



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""

from typing import Optional, Protocol

# MyPy cannot resolve this during pull request
from pdm.project.config import ConfigItem as _ConfigItem  # type: ignore

from .plugin import BumpCommand as _Command


class _CoreLike(Protocol):
    """"""

    def register_command(
        self, command: type[_Command], name: Optional[str] = None
    ) -> None:
        """

        Parameters
        ----------
        command: type[_Command] :

        name: Optional[str] :
             (Default value = None)

        Returns
        -------

        """
        # Method empty: Only a protocol stub
        raise NotImplementedError()

    @staticmethod
    def add_config(name: str, config_item: _ConfigItem) -> None:
        """

        Parameters
        ----------
        name: str :

        config_item: _ConfigItem :


        Returns
        -------

        """
        # Method empty: Only a protocol stub
        raise NotImplementedError()


def main(core: _CoreLike) -> None:
    """

    Parameters
    ----------
    core: _CoreLike :


    Returns
    -------

    """
    core.register_command(_Command)



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from argparse import ArgumentParser
from traceback import format_exc as get_traceback
from typing import final

from ..core.logging import logger
from ..core.version import Version
from .base import VersionModifier, VersionPersister, action


@final
@action
class SetExplicitVersionModifier(VersionModifier):
    """"""

    name: str = "to"
    description: str = "Sets a new version explicitly from CLI"

    def __init__(
        self,
        version: Version,
        persister: VersionPersister,
        new_version: list[str],
        **kwargs,
    ) -> None:
        super().__init__(version, persister, **kwargs)
        if len(new_version) != 1:
            raise ValueError("Only one value is supported")
        self.__new_version: str = new_version[0]

    def create_new_version(self) -> Version:
        """"""
        try:
            new_version: Version = Version.from_string(self.__new_version)
            return new_version
        except ValueError as exc:
            err: str = f"'{self.__new_version}' is not a valid value."
            logger.exception(err, exc_info=True)
            logger.debug("Exception occurred: %s", get_traceback())
            raise ValueError(err) from exc

    @classmethod
    def _update_command(cls, sub_parser: ArgumentParser) -> None:
        """

        Parameters
        ----------
        sub_parser: ArgumentParser :


        Returns
        -------

        """
        sub_parser.add_argument(
            "new_version",
            action="store",
            nargs=1,
            type=str,
            default=None,
            help="The next version to set.",
        )

        VersionModifier._update_command(sub_parser)



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel, Chase Sterling.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from functools import cached_property
from pathlib import Path
from re import M as MultilinePattern  # noqa: N811
from re import Match, Pattern
from re import compile as compile_re
from typing import Final, Optional, cast

from .core.config import Config, ConfigKeys, ConfigValues
from .core.logging import logger
from .core.version import Pep440VersionFormatter, Version

DEFAULT_REGEX: Final[Pattern[str]] = compile_re(
    r"^__version__\s*=\s*[\"'](?P<version>.+?)[\"']\s*(?:#.*)?$",
    MultilinePattern,
)


class DynamicVersionConfig:
    """"""

    __file: Path
    __pattern: Pattern[str]
    __file_encoding: str

    def __init__(
        self, file_path: Path, pattern: Pattern[str], encoding: str = "utf-8"
    ) -> None:
        self.__file = file_path
        self.__pattern = pattern
        self.__file_encoding = encoding

    @property
    def file(self) -> Path:
        """"""
        return self.__file

    @property
    def pattern(self) -> Pattern[str]:
        """"""
        return self.__pattern

    @cached_property
    def dynamic_version(self) -> Optional[str]:
        """"""
        with self.__file.open("r", encoding=self.__file_encoding) as file_ptr:
            match = self.__pattern.search(file_ptr.read())
        if match is not None:
            return match.group(ConfigKeys.VERSION)
        return None

    def replace_dynamic_version(self, new_version: str) -> None:
        """

        Parameters
        ----------
        new_version: str :


        Returns
        -------

        """
        with self.__file.open("r", encoding=self.__file_encoding) as file_ptr:
            version_file = file_ptr.read()
            match = self.__pattern.search(version_file)
            if match is None:
                raise ValueError("Failed to fetch version")
            match = cast(Match[str], match)
            version_start, version_end = match.span(ConfigKeys.VERSION)
            new_version_file = (
                version_file[:version_start]
                + new_version
                + version_file[version_end:]
            )
        with self.__file.open("w", encoding=self.__file_encoding) as file_ptr:
            file_ptr.write(new_version_file)

    @staticmethod
    def find_dynamic_config(
        root_path: Path, project_config: Config
    ) -> Optional["DynamicVersionConfig"]:
        """

        Parameters
        ----------
        root_path: Path :

        project_config: Config :


        Returns
        -------

        """
        if (
            project_config.get_pyproject_build_system(ConfigKeys.BUILD_BACKEND)
            == ConfigValues.DEPREACTED_BUILD_BACKEND_PDM_PEP517_API
            and project_config.get_pyproject_tool_config(
                ConfigKeys.VERSION, ConfigKeys.VERSION_SOURCE
            )
            == ConfigValues.VERSION_SOURCE_FILE
        ):
            logger.warning(
                "Build backend pdm-pep517 is deprecated. Consider upgrading."
            )
            file_path = project_config.get_pyproject_tool_config(
                ConfigKeys.VERSION, ConfigKeys.VERSION_SOURCE_FILE_PATH
            )
            return DynamicVersionConfig(
                file_path=root_path / file_path,
                pattern=DEFAULT_REGEX,
            )
        if (
            project_config.get_pyproject_build_system(ConfigKeys.BUILD_BACKEND)
            == ConfigValues.BUILD_BACKEND_PDM_BACKEND
        ):
            if (
                project_config.get_pyproject_tool_config(
                    ConfigKeys.VERSION, ConfigKeys.VERSION_SOURCE
                )
                == ConfigValues.VERSION_SOURCE_SCM
            ):
                logger.error(
                    "PDM bump cannot be used if version is fetched from scm"
                )
            if (
                project_config.get_pyproject_tool_config(
                    ConfigKeys.VERSION, ConfigKeys.VERSION_SOURCE
                )
                == ConfigValues.VERSION_SOURCE_FILE
            ):
                file_path = project_config.get_pyproject_tool_config(
                    ConfigKeys.VERSION, ConfigKeys.VERSION_SOURCE_FILE_PATH
                )
                return DynamicVersionConfig(
                    file_path=root_path / file_path,
                    pattern=DEFAULT_REGEX,
                )
        return None


# Justification: Implementation of minimal protocol
class DynamicVersionSource:  # pylint: disable=R0903
    """"""

    def __init__(self, project_root: Path, config: Config) -> None:
        self.__project_root = project_root
        self.__config = config
        self.__current_version: Optional[Version] = None

    @property
    def is_enabled(self) -> bool:
        """"""
        return self.__config.meta_data.is_dynamic_version

    def __get_current_version(self) -> Version:
        if self.__current_version is not None:
            return cast(Version, self.__current_version)

        dynamic: DynamicVersionConfig = self.__get_dynamic_version()
        version: Optional[str] = dynamic.dynamic_version
        if version is not None:
            self.__current_version = Version.from_string(version)
            return self.__current_version
        raise ValueError(
            f"Failed to find version in {dynamic.file}. "
            f"Make sure it matches {dynamic.pattern}"
        )

    def __set_current_version(self, version: Version) -> None:
        self.__current_version = version
        ver: str = Pep440VersionFormatter().format(version)
        config: DynamicVersionConfig = self.__get_dynamic_version()
        config.replace_dynamic_version(ver)

    current_version = property(__get_current_version, __set_current_version)

    def __get_dynamic_version(self) -> DynamicVersionConfig:
        dynamic_version: Optional[
            DynamicVersionConfig
        ] = DynamicVersionConfig.find_dynamic_config(
            self.__project_root, self.__config
        )
        if dynamic_version is not None:
            dynamic: DynamicVersionConfig = cast(
                DynamicVersionConfig, dynamic_version
            )
            return dynamic
        raise ValueError(
            f"Failed to extract dynamic version from {self.__project_root}."
            f" Only pdm-pep517 `file` types are supported."
        )



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import InitVar, dataclass, field
from enum import Enum, auto
from functools import cached_property
from re import Match, Pattern, compile
from typing import Callable, Final, Mapping, Optional

from ..core.logging import logger
from ..core.version import NonNegativeInteger

""""""


class _StrEnum(str, Enum):
    pass


class CommitType(_StrEnum):
    """"""

    Undefined = auto()
    Feature = "feat"
    Bugfix = "fix"
    Chore = "chore"
    QualityAssurance = "qa"
    Documentation = "docs"
    Build = "build"
    ContinuousIntegration = "ci"
    CodeStyle = "style"
    Refactoring = "refactor"
    Performance = "perf"
    Test = "test"

    @classmethod
    def _missing_(cls, value: object) -> Enum:
        return CommitType.Undefined

    @classmethod
    def all_values(cls) -> Mapping["CommitType", str]:
        mapping: dict["CommitType", str] = {}
        base_items: list[str] = dir(_StrEnum)
        for key in dir(cls):
            if (
                not key.startswith("_")
                and key != "all_values"
                and key not in base_items
            ):
                mapping[cls[key]] = cls[key].value

        return mapping


class CommitParser(ABC):
    """"""

    @abstractmethod
    def is_breaking_change(self, commit: "Commit") -> bool:
        """"""
        raise NotImplementedError()

    @abstractmethod
    def parse_commit_type(self, commit: "Commit") -> CommitType:
        """"""
        raise NotImplementedError()


_DEFAULT_MAPPINGS: Mapping[CommitType, str] = CommitType.all_values()


class ConventionalCommitParser(CommitParser):
    """"""

    def __init__(
        self, prefix_mappings: Mapping[CommitType, str] = _DEFAULT_MAPPINGS
    ) -> None:
        self.__line_parser: Pattern = compile(r"^([^:!]+!?):\s.*", flags=0)
        self.__prefix_mappings: Mapping[CommitType, str] = prefix_mappings

    def is_breaking_change(self, commit: "Commit") -> bool:
        """"""
        breaking_change_suffix = "!"

        prefix: Optional[str] = self.__get_conventional_commit_prefix(commit)
        if prefix is not None:
            return prefix.endswith(breaking_change_suffix)

        return False

    def parse_commit_type(self, commit: "Commit") -> CommitType:
        """"""
        prefix: Optional[str] = self.__get_conventional_commit_prefix(commit)
        if prefix is not None:
            for prefix_to_commit_type in self.__prefix_mappings.items():
                commit_type, configured_prefix = prefix_to_commit_type
                if prefix.startswith(configured_prefix):
                    return commit_type

        return CommitType.Undefined

    def __get_conventional_commit_prefix(
        self, commit: "Commit"
    ) -> Optional[str]:
        _match: Optional[Match[str]] = self.__line_parser.match(commit.header)
        if _match is not None:
            for group in _match.groups():
                logger.debug(
                    "Found prefix string '%s' for commit '%s'",
                    group,
                    commit.header,
                )
                return group

        return None


@dataclass(init=True, eq=True, order=True, repr=True)
class Commit:
    """"""

    header: Final[str] = field(init=True, repr=True, hash=True, compare=True)
    _commit_parser: "CommitParser" = field(
        init=False, compare=False, hash=False, repr=True
    )
    commit_parser_factory: InitVar[
        Optional[Callable[[], "CommitParser"]]
    ] = None  # NOSONAR

    def __post_init__(
        self,
        commit_parser_factory: Optional[Callable[[], "CommitParser"]] = None,
    ) -> None:
        """"""
        # Justification: False positive
        self._commit_parser = (
            commit_parser_factory()
            if commit_parser_factory is not None
            else ConventionalCommitParser()
        )

    @cached_property
    def commit_type(self) -> CommitType:
        """"""
        return self._commit_parser.parse_commit_type(self)

    @cached_property
    def is_breaking_change(self) -> bool:
        """"""
        return self._commit_parser.is_breaking_change(self)


@dataclass(frozen=True, repr=True)
class CommitStatistics:
    """"""

    commit_type_count: Mapping[CommitType, NonNegativeInteger] = field(
        init=True, default_factory=dict
    )
    contains_breaking_changes: bool = field(init=True, default=False)


@dataclass(init=True, eq=True, order=True, repr=True, frozen=True)
class History:
    """"""

    commits: Iterable[Commit] = field(init=True, default_factory=list)

    @cached_property
    def get_commit_stats(self) -> CommitStatistics:
        """"""
        items: dict[CommitType, NonNegativeInteger] = {}

        breaking_changes = False
        for commit in self.commits:
            logger.debug(
                "Commit '%s' is of type %s",
                commit.header,
                commit.commit_type.name,
            )
            commit_type: CommitType = commit.commit_type
            items[commit_type] = (
                1 if commit_type not in items else items[commit_type] + 1
            )
            breaking_changes = breaking_changes or commit.is_breaking_change
            if commit.is_breaking_change:
                logger.debug("Commit '%s' is breaking change", commit.header)

        logger.debug(
            "Found the following commit history: \n%s",
            self.__format_for_debug(),
        )

        return CommitStatistics(
            commit_type_count=items, contains_breaking_changes=breaking_changes
        )

    def __format_for_debug(self) -> str:
        """"""
        lines = []

        lines.append("[B] | Type      | Commit     ")
        lines.append("-----------------------------")

        for commit in self.commits:
            b = " X " if commit.is_breaking_change else "   "
            c_t = commit.commit_type.value
            lines.append(f"{b} | {c_t: <9} | {commit.header} ")

        return "\n".join(lines)



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from argparse import ArgumentParser, Namespace
from pathlib import Path
from traceback import format_exc as get_traceback
from typing import Final, Optional, Protocol, cast, final

# MyPy does not recognize this during pull requests
from pdm.cli.commands.base import BaseCommand  # type: ignore
from pdm.termui import UI  # type: ignore

from .actions import actions
from .core.config import Config, ConfigHolder, ConfigKeys, ConfigSections
from .core.logging import (
    logger,
    setup_logger,
    traced_function,
    update_logger_from_project_ui,
)
from .core.version import Pep440VersionFormatter, Version
from .dynamic import DynamicVersionSource
from .source import StaticPep621VersionSource
from .vcs import (
    DefaultVcsProvider,
    VcsProvider,
    VcsProviderRegistry,
    vcs_providers,
)


# Justification: Protocol for interoperability
class _CoreLike(Protocol):  # pylint: disable=R0903
    """"""

    ui: UI


class _ProjectLike(ConfigHolder, Protocol):
    """"""

    root: Path
    core: _CoreLike
    PYPROJECT_FILENAME: str

    def write_pyproject(self, show_message: bool) -> None:
        """

        Parameters
        ----------
        show_message: bool :


        Returns
        -------

        """
        # Method empty: Only a protocol stub
        raise NotImplementedError()


# Justification: Minimal protocol. Maybe false positive,
# since 2 public methods available
class _VersionSource(Protocol):  # pylint: disable=R0903
    """"""

    @property
    def is_enabled(self) -> bool:
        """"""
        raise NotImplementedError()

    def __get_current_version(self) -> Version:
        raise NotImplementedError()

    def __set_current_version(self, version: Version) -> None:
        raise NotImplementedError()

    current_version: property = property(
        __get_current_version, __set_current_version
    )


@final
class BumpCommand(BaseCommand):
    """"""

    name: Final[str] = "bump"
    description: str = "Bumps the version to a next version following PEP440."

    def __init__(self) -> None:
        super().__init__()
        self.__backend: Optional[_VersionSource] = None

    @traced_function
    def add_arguments(self, parser: ArgumentParser) -> None:
        """

        Parameters
        ----------
        parser: ArgumentParser :


        Returns
        -------

        """
        actions.update_parser(parser)

    @traced_function
    def save_version(self, version: Version) -> None:
        """

        Parameters
        ----------
        version: Version :


        Returns
        -------

        """
        if self.__backend is None:
            msg = ". ".join(
                (
                    "No source for a version could be determined.",
                    "Saving version failed.",
                )
            )
            logger.error(msg)
            return

        self.__backend.current_version = version

    @traced_function
    def handle(self, project: _ProjectLike, options: Namespace) -> None:
        """

        Parameters
        ----------
        project: _ProjectLike :

        options: Namespace :


        Returns
        -------

        """
        # This will not handling tracing or related parts
        # Should be evaluated at start-up time
        if hasattr(options, "verbose"):
            setup_logger(options.verbose)

        config: Config = Config(project)
        update_logger_from_project_ui(project.core.ui)

        selected_backend: Optional[_VersionSource] = self._select_backend(
            project, config
        )

        self.__backend = selected_backend

        if selected_backend is None:
            pyproject_file = project.root / project.PYPROJECT_FILENAME
            logger.error("Cannot find version in %s", pyproject_file)
            return

        backend: _VersionSource = cast(_VersionSource, selected_backend)
        vcs_provider: VcsProvider = self._get_vcs_provider(project)

        try:
            actions.execute(
                options,
                version=backend.current_version,
                persister=self,
                vcs_provider=vcs_provider,
            )
        except ValueError as exc:
            logger.exception("Failed to execute action", exc_info=True)
            logger.debug("Exception occurred: %s", get_traceback())
            raise SystemExit(1) from exc

    @traced_function
    def _get_vcs_provider(self, project: _ProjectLike) -> VcsProvider:
        """

        Parameters
        ----------
        project: _ProjectLike :


        Returns
        -------

        """
        config: Config = Config(project)
        value = config.get_config_or_pyproject_value(
            ConfigSections.PDM_BUMP,
            ConfigSections.PDM_BUMP_VCS,
            ConfigKeys.VCS_PROVIDER,
        )

        registry: VcsProviderRegistry = vcs_providers

        if value is not None:
            provider_name: str = str(value)
            factory_method = registry[provider_name]
            if factory_method is not None:
                factory = factory_method()
                return factory.force_create_provider(project.root)
        else:
            provider = registry.find_repository_root(project.root)
            if provider is not None:
                return provider

        return DefaultVcsProvider(project.root)

    @traced_function
    def _select_backend(
        self, project: _ProjectLike, config: Config
    ) -> Optional[_VersionSource]:
        """

        Parameters
        ----------
        project: _ProjectLike :

        config: Config :


        Returns
        -------

        """
        static_backend: _VersionSource = StaticPep621VersionSource(
            project, config
        )
        dynamic_backend: _VersionSource = DynamicVersionSource(
            project.root, config
        )

        selected_backend: Optional[_VersionSource] = None
        for backend in (static_backend, dynamic_backend):
            if backend.is_enabled:
                selected_backend = backend
                break

        return selected_backend

    @traced_function
    def _version_to_string(self, version: Version) -> str:
        """

        Parameters
        ----------
        version: Version :


        Returns
        -------

        """
        result: str = Pep440VersionFormatter().format(version)
        return result



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""

from typing import Literal, Optional, final

from ..core.logging import logger, traced_function
from ..core.version import Version
from .base import VersionModifier, action

# Comparable functions at poetry. Cf.
# https://python-poetry.org/docs/cli/#version


@final
@action
class PoetryLikePreMajorVersionModifier(VersionModifier):
    """"""

    name: str = "premajor"
    description: str = (
        "Prepares a new major version - like `poetry version premajor`"
    )

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        if not self.current_version.is_final:
            logger.error(
                "Cannot create a new major pre-release "
                + "from non-final version %s",
                self.current_version,
            )
            raise ValueError(self.current_version)

        major_version = self.current_version.major + 1
        release_part = (major_version, 0, 0)
        alpha_part = 0

        next_version: Version = Version(
            epoch=self.current_version.epoch,
            release_tuple=release_part,
            preview=("a", alpha_part),
            dev=None,
            local=None,
            post=None,
        )

        return next_version


@final
@action
class PoetryLikePreMinorVersionModifier(VersionModifier):
    """"""

    name: str = "preminor"
    description: str = (
        "Prepares a new minor version - like `poetry version preminor`"
    )

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        if not self.current_version.is_final:
            logger.error(
                "Cannot create a new minor pre-release "
                + "from non-final version %s",
                self.current_version,
            )
            raise ValueError(self.current_version)

        minor_version = self.current_version.minor + 1
        release_part = (self.current_version.major, minor_version, 0)
        alpha_part = 0

        next_version: Version = Version(
            epoch=self.current_version.epoch,
            release_tuple=release_part,
            preview=("a", alpha_part),
            dev=None,
            local=None,
            post=None,
        )

        return next_version


@final
@action
class PoetryLikePrePatchVersionModifier(VersionModifier):
    """"""

    name: str = "prepatch"
    description: str = (
        "Prepares a new patch (micro) version - like `poetry version prepatch`"
    )

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        if not self.current_version.is_final:
            logger.error(
                "Cannot create a new patch pre-release "
                + "from non-final version %s",
                self.current_version,
            )
            raise ValueError(self.current_version)

        micro_version = self.current_version.micro + 1
        release_part = (
            self.current_version.major,
            self.current_version.minor,
            micro_version,
        )
        alpha_part = 0

        next_version: Version = Version(
            epoch=self.current_version.epoch,
            release_tuple=release_part,
            preview=("a", alpha_part),
            dev=None,
            local=None,
            post=None,
        )

        return next_version


@final
@action
class PoetryLikePreReleaseVersionModifier(VersionModifier):
    """"""

    name: str = "prerelease"
    description: str = (
        "Prepares a new release version - like `poetry version prerelease`"
    )

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        release_part = (
            self.current_version.major,
            self.current_version.minor,
            self.current_version.micro,
        )
        preview_part: Optional[
            tuple[Literal["a", "b", "c", "alpha", "beta", "rc"], int]
        ] = None

        if self.current_version.preview is not None:
            if self.current_version.is_alpha:
                preview_part = ("a", self.current_version.preview[1] + 1)
            elif self.current_version.is_beta:
                preview_part = ("b", self.current_version.preview[1] + 1)
            elif self.current_version.is_release_candidate:
                preview_part = ("rc", self.current_version.preview[1] + 1)
        else:
            preview_part = ("a", 0)
            release_part = (
                self.current_version.major,
                self.current_version.minor,
                self.current_version.micro + 1,
            )

        next_version: Version = Version(
            epoch=self.current_version.epoch,
            release_tuple=release_part,
            preview=preview_part,
            post=None,
            dev=None,
            local=None,
        )

        return next_version



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from abc import abstractmethod
from functools import cached_property
from typing import Optional, final

from ..core.logging import logger, silenced
from ..core.version import Pep440VersionFormatter, Version
from ..vcs import (
    CommitStatistics,
    CommitType,
    History,
    VcsProvider,
    VcsProviderAggregator,
)
from .base import VersionConsumer, VersionModifier, VersionPersister, action
from .version_providers import SemanticVersionPolicy, VersionPolicy


@final
@action
class CreateTagFromVersion(VersionConsumer, VcsProviderAggregator):
    """"""

    name: str = "tag"
    description: str = "Create a VCS revision tag from the current version"

    def __init__(
        self, version: Version, vcs_provider: VcsProvider, **kwargs
    ) -> None:
        VersionConsumer.__init__(self, version, **kwargs)
        VcsProviderAggregator.__init__(self, vcs_provider, **kwargs)

    def run(self, dry_run: bool = False) -> None:
        """

        Parameters
        ----------
        dry_run: bool :
             (Default value = False)

        Returns
        -------

        """
        if not dry_run:
            self.vcs_provider.create_tag_from_version(self.current_version)
        else:
            logger.info(
                "Would create tag v%s",
                Pep440VersionFormatter().format(self.current_version),
            )

    @classmethod
    def get_allowed_arguments(cls) -> set[str]:
        """"""
        return set(["vcs_provider"]).union(
            VersionConsumer.get_allowed_arguments()
        )


class _VcsVersionDerivatingVersionConsumer(
    VersionConsumer, VcsProviderAggregator
):
    """"""

    def __init__(
        self, version: Version, vcs_provider: VcsProvider, **kwargs
    ) -> None:
        """"""
        VersionConsumer.__init__(self, version, **kwargs)
        VcsProviderAggregator.__init__(self, vcs_provider, **kwargs)

    @classmethod
    def get_allowed_arguments(cls) -> set[str]:
        """"""
        return set(["vcs_provider"]).union(
            VersionConsumer.get_allowed_arguments()
        )

    def save_version(self, _: Version) -> None:
        """"""
        # This does not persist anything
        pass  # pylint: disable=W0107

    def derive_next_version(self) -> Optional[Version]:
        """"""
        history: History = self.vcs_provider.get_history()
        stats: CommitStatistics = history.get_commit_stats

        keys: frozenset[CommitType] = frozenset(stats.commit_type_count.keys())

        if len(keys) == 0:
            logger.info("History clean. No need to update version")
            return None

        modifier: VersionModifier = self._version_policy.get_modifier(
            stats, self.vcs_provider.is_clean
        )

        with silenced(logger):
            new_version: Version = modifier.create_new_version()
            return new_version

    @property
    @abstractmethod
    def _version_policy(self) -> VersionPolicy:
        """"""
        raise NotImplementedError()


@final
@action
class SuggestNewVersion(_VcsVersionDerivatingVersionConsumer):
    """"""

    name: str = "suggest"
    description: str = "Suggests a new version from the VCS"

    def run(self, dry_run: bool = False) -> None:
        """"""
        logger.debug("Ignoring dry run parameter set to %s", dry_run)
        new_version: Optional[Version] = self.derive_next_version()
        if new_version is not None:
            next_version: str = Pep440VersionFormatter().format(new_version)
            logger.info("Would suggest new version: %s", next_version)

    @cached_property
    def _version_policy(self) -> VersionPolicy:
        return SemanticVersionPolicy(self.current_version)


@final
# when implemented, add @action
class AutoSelectVersionModifier(
    VersionModifier, _VcsVersionDerivatingVersionConsumer
):
    """"""

    name: str = "auto"
    description: str = "Create a new version from the VCS"

    def __init__(
        self,
        version: Version,
        persister: VersionPersister,
        vcs_provider: VcsProvider,
        **kwargs,
    ) -> None:
        """"""
        VersionModifier.__init__(self, version, persister, **kwargs)
        _VcsVersionDerivatingVersionConsumer.__init__(
            self, version, vcs_provider, **kwargs
        )

    def create_new_version(self) -> Version:
        """"""
        raise NotImplementedError

    @cached_property
    def _version_policy(self) -> VersionPolicy:
        raise NotImplementedError()



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from functools import cached_property
from pathlib import Path
from subprocess import CalledProcessError
from typing import Optional, cast

from ..core.logging import logger
from ..core.version import Pep440VersionFormatter, Version
from .core import VcsProvider, VcsProviderError, vcs_provider
from .git import GitCommonVcsProviderFactory
from .history import Commit, History
from .mixins import CliRunnerMixin


class GitCliVcsProvider(VcsProvider, CliRunnerMixin):
    """"""

    __GIT_EXECUTABLE_NAME = "git"

    @cached_property
    def git_executable_path(self) -> Path:
        """"""
        logger.debug("Searching for %s executable", self.__GIT_EXECUTABLE_NAME)
        git_executable_path: Optional[Path] = self._which(
            self.__GIT_EXECUTABLE_NAME
        )
        if git_executable_path is None:
            raise FileNotFoundError(
                f"No executable '{self.__GIT_EXECUTABLE_NAME}' found in PATH"
            )
        return cast(Path, git_executable_path)

    @property
    def is_available(self) -> bool:
        """"""
        logger.debug("Checking, if we're working on a git repository")
        exit_code, _out, _err = self.run(
            self.git_executable_path,
            ("rev-parse", "--root-dir"),
            raise_on_exit=False,
            cwd=self.current_path,
        )
        return 0 == exit_code

    @property
    def is_clean(self) -> bool:
        """"""
        try:
            logger.debug("Checking, if there are any changes")
            _ex, out, _err = self.run(
                self.git_executable_path,
                ("status", "--porcelain"),
                cwd=self.current_path,
                raise_on_exit=True,
            )

            dirty_files = [
                ln
                for ln in out.splitlines()
                if not ln.strip().startswith("??")
            ]

            logger.debug(
                "The following files have been reported as modified: \n %s",
                "\n".join(dirty_files),
            )

            return len(dirty_files) == 0
        except CalledProcessError as cpe:
            raise VcsProviderError(
                f"Failed to check if {self.current_path} is a clean "
                f"git repository. Reason: {cpe.stderr}"
            ) from cpe

    def check_in_items(self, message: str, *files: tuple[Path, ...]) -> None:
        """

        Parameters
        ----------
        message: str :

        *files: tuple[Path, ...] :


        Returns
        -------

        """
        try:
            args: list[str] = ["add", "--update"]
            args.extend(str(f) for f in files)
            logger.debug(
                "Checking in the following files: \n%s",
                ", ".join([str(f) for f in files]),
            )
            _ = self.run(
                self.git_executable_path,
                tuple(args),
                raise_on_exit=True,
                cwd=self.current_path,
            )
            logger.debug("Creating new commit with message '%s'", message)
            _ = self.run(
                self.git_executable_path,
                ("commit", "-m", f'"{message}"'),
                raise_on_exit=True,
                cwd=self.current_path,
            )
        except CalledProcessError as cpe:
            f_args = ",".join([str(f) for f in files])
            raise VcsProviderError(
                f"Failed to check in items in {self.current_path} "
                f"using {f_args}. Reason: {cpe.stderr}"
            ) from cpe

    def create_tag_from_string(self, version_formatted: str) -> None:
        """

        Parameters
        ----------
        version_formatted: str :


        Returns
        -------

        """
        try:
            logger.info(
                "Creating tag '%s' on current commit.", version_formatted
            )
            _ = self.run(
                self.git_executable_path,
                ("tag", version_formatted),
                raise_on_exit=True,
                cwd=self.current_path,
            )
        except CalledProcessError as cpe:
            raise VcsProviderError(
                f"Failed to create tag {version_formatted} "
                f"in {self.current_path}. Reason: {cpe.stderr}"
            ) from cpe

    def get_most_recent_tag(self) -> Optional[Version]:
        """"""
        try:
            logger.debug("Checking for most recent tag.")
            _, output, _ = self.run(
                self.git_executable_path,
                ("describe", "--tags", "--abbrev=0"),
                raise_on_exit=True,
                cwd=self.current_path,
            )
        except CalledProcessError as cpe:
            raise VcsProviderError(
                f"Failed to receive most recent tag. . Reason: {cpe.stderr}"
            ) from cpe

        if output.strip() == "":
            logger.debug("Could not find a tagged version.")
            return None

        logger.debug("Found version '%s'", output)
        return Version.from_string(output)

    def get_number_of_changes_since_last_release(self) -> int:
        """"""
        try:
            logger.debug("Searching last tag ...")
            _, output, _ = self.run(
                self.git_executable_path,
                ("describe", "--tags"),
                raise_on_exit=False,
                cwd=self.current_path,
            )
            if output.strip() == "":
                raise VcsProviderError("Failed to fetch most recent tag")
            logger.debug("Found tag '%s'", output)
            _, output, _ = self.run(
                self.git_executable_path,
                ("rev-list", f"{output}..HEAD", "--count"),
                raise_on_exit=True,
                cwd=self.current_path,
            )
            logger.debug("Git return %s changes", output)

            return int(output)
        except CalledProcessError as cpe:
            raise VcsProviderError(
                f"Failed to receive number of commits since last tag."
                f" Reason: {cpe.stderr}"
            ) from cpe

    def get_changes_not_checked_in(self) -> int:
        """"""
        try:
            logger.debug("Searching for changes")
            _, output, _ = self.run(
                self.git_executable_path,
                ("status", "--porcelain"),
                raise_on_exit=True,
                cwd=self.current_path,
            )
            logger.debug(
                "The following have been modified or added:\n%s", output
            )
            lines = output.split("\n")
            return len(lines)
        except CalledProcessError as cpe:
            raise VcsProviderError(
                f"Failed to receive number of changes that are not committed."
                f" Reason: {cpe.stderr}"
            ) from cpe

    def get_history(self, since_last_tag: bool = True) -> History:
        """"""
        commit_history: str = ""
        if since_last_tag:
            last_tag: Optional[Version] = self.get_most_recent_tag()
            if last_tag is not None:
                last_tag_name: str = (
                    f"v{Pep440VersionFormatter().format(last_tag)}"
                )
                commit_history = f"{last_tag_name}..HEAD"

        try:
            _, output, _ = self.run(
                self.git_executable_path,
                ("log", commit_history, "--format=%s"),
                raise_on_exit=True,
                cwd=self.current_path,
            )
            logger.debug(
                "The following commits have been received for %s:\n%s",
                commit_history,
                output,
            )
            commit_texts: list[str] = output.split("\n")

            commits: list[Commit] = [
                Commit(t) for t in commit_texts if t.strip() != ""
            ]

            return History(commits)

        except CalledProcessError as cpe:
            raise VcsProviderError(
                f"Failed to receive commit history. Reason: {cpe.stderr}"
            ) from cpe


@vcs_provider("git-cli")
class GitCliVcsProviderFactory(GitCommonVcsProviderFactory):
    """"""

    def _create_provider(self, path: Path) -> VcsProvider:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        return GitCliVcsProvider(path)



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from abc import ABC, abstractmethod
from collections.abc import Iterator
from pathlib import Path
from typing import AnyStr, Callable, Final, NamedTuple, Optional, Union, cast

from ..core.version import Pep440VersionFormatter, Version
from .history import History

_Pathlike = Union[Path, AnyStr]


class _PathLikeConverter(ABC):
    """"""

    @staticmethod
    def _pathlike_to_path(path: _Pathlike) -> Path:
        """

        Parameters
        ----------
        path: _Pathlike :


        Returns
        -------

        """
        if isinstance(path, Path):
            return cast(Path, path)
        if isinstance(path, str):
            return Path(path)
        if isinstance(path, bytes):
            return Path(cast(bytes, path).decode("utf-8"))

        raise ValueError(
            f"'{path}' must be a valid path, string or bytes instance"
        )


class VcsFileSystemIdentifier(NamedTuple):
    """"""

    file_name: Optional[str]
    dir_name: Optional[str]


class VcsProvider(_PathLikeConverter, ABC):
    """"""

    def __init__(self, path: _Pathlike) -> None:
        self.__path = _PathLikeConverter._pathlike_to_path(path)

    @property
    def is_available(self) -> bool:
        """"""
        return False

    @property
    def current_path(self) -> Path:
        """"""
        return self.__path

    @property
    @abstractmethod
    def is_clean(self) -> bool:
        """"""
        raise NotImplementedError()

    @abstractmethod
    def check_in_items(self, message: str, *files: tuple[Path, ...]) -> None:
        """

        Parameters
        ----------
        message: str :

        *files: tuple[Path, ...] :


        Returns
        -------

        """
        raise NotImplementedError()

    def create_tag_from_version(
        self, version: Version, prepend_letter_v: bool = True
    ) -> None:
        """

        Parameters
        ----------
        version: Version :

        prepend_letter_v: bool :
             (Default value = True)

        Returns
        -------

        """
        version_formatted: str = Pep440VersionFormatter().format(version)
        if prepend_letter_v:
            version_formatted = "v" + version_formatted
        self.create_tag_from_string(version_formatted)

    @abstractmethod
    def create_tag_from_string(self, version_formatted: str) -> None:
        """

        Parameters
        ----------
        version_formatted: str :


        Returns
        -------

        """
        raise NotImplementedError()

    @abstractmethod
    def get_most_recent_tag(self) -> Optional[Version]:
        """"""
        raise NotImplementedError()

    @abstractmethod
    def get_number_of_changes_since_last_release(self) -> int:
        """"""
        raise NotImplementedError()

    @abstractmethod
    def get_changes_not_checked_in(self) -> int:
        """"""
        raise NotImplementedError()

    @abstractmethod
    def get_history(self, since_last_tag: bool = True) -> History:
        """"""
        raise NotImplementedError()


class VcsProviderAggregator:
    """"""

    def __init__(self, vcs_provider: VcsProvider, **kwargs) -> None:
        self.__vcs_provider: VcsProvider = vcs_provider

    @property
    def vcs_provider(self) -> VcsProvider:
        """"""
        return self.__vcs_provider


class VcsProviderFactory(_PathLikeConverter, ABC):
    """"""

    def force_create_provider(self, path: Path) -> VcsProvider:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        return self._create_provider(path)

    @abstractmethod
    def _create_provider(self, path: Path) -> VcsProvider:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def vcs_fs_root(self) -> Iterator[VcsFileSystemIdentifier]:
        """"""
        raise NotImplementedError()

    def find_repository_root(self, path: _Pathlike) -> Optional[VcsProvider]:
        """

        Parameters
        ----------
        path: _Pathlike :


        Returns
        -------

        """
        real_path: Path = _PathLikeConverter._pathlike_to_path(path)
        return self.find_repository_root_from_path(real_path)

    def find_repository_root_from_path(
        self, path: Path
    ) -> Optional[VcsProvider]:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        if not path.is_dir():
            raise ValueError(f"{path} must refer to a directory.")

        cur_path: Path = path
        while len(cur_path.parts) > 1:  # First part refers to root level
            if self._is_valid_root(cur_path):
                return self._create_provider(cur_path)
            cur_path = cur_path.parent

        return None

    def _is_valid_fs_root_file(self, path: Path) -> bool:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        return path is not None and path.exists() and path.is_file()

    def _is_valid_fs_root_dir(self, path: Path) -> bool:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        return path is not None and path.exists() and path.is_dir()

    def _directory_exists(self, path: Path, directory_name: str) -> bool:
        """

        Parameters
        ----------
        path: Path :

        directory_name: str :


        Returns
        -------

        """
        dir_path: Path = path / directory_name
        return self._is_valid_fs_root_dir(dir_path)

    def _file_exists(self, path: Path, file_name: str) -> bool:
        """

        Parameters
        ----------
        path: Path :

        file_name: str :


        Returns
        -------

        """
        file_path: Path = path / file_name
        return self._is_valid_fs_root_file(file_path)

    def _is_valid_root(self, path: Path) -> bool:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        is_repository_root = False
        for fs_root in self.vcs_fs_root:
            is_valid_root_by_file: bool = (
                fs_root.file_name is None
                or self._file_exists(path, fs_root.file_name)
            )
            is_valid_root_by_dir: bool = (
                fs_root.dir_name is None
                or self._directory_exists(path, fs_root.dir_name)
            )
            is_repository_root = is_repository_root or (
                is_valid_root_by_dir and is_valid_root_by_file
            )

        return is_repository_root


class VcsProviderRegistry(
    dict[str, Callable[..., VcsProviderFactory]], _PathLikeConverter
):
    """"""

    def find_repository_root(self, path: _Pathlike) -> Optional[VcsProvider]:
        """

        Parameters
        ----------
        path: _Pathlike :


        Returns
        -------

        """
        real_path: Path = _PathLikeConverter._pathlike_to_path(path)
        return self.find_repository_root_by_path(real_path)

    def find_repository_root_by_path(
        self, path: Path
    ) -> Optional[VcsProvider]:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        for _, value in self.items():
            factory: VcsProviderFactory = value()
            result: Optional[
                VcsProvider
            ] = factory.find_repository_root_from_path(path)
            if result is not None:
                return result

        return None

    def register(self, name: str) -> Callable:
        """

        Parameters
        ----------
        name: str :


        Returns
        -------

        """

        def decorator(clazz: type[VcsProvider]):
            """

            Parameters
            ----------
            clazz: type[VcsProvider] :


            Returns
            -------

            """
            if not issubclass(clazz, VcsProviderFactory):
                raise ValueError(
                    f"{clazz.__name__} is not an sub-type of "
                    f"{VcsProviderFactory.__name__}"
                )

            self[name] = clazz

        return decorator

    def __missing__(
        self, key: str
    ) -> Optional[Callable[..., VcsProviderFactory]]:
        return None


vcs_providers: Final[VcsProviderRegistry] = VcsProviderRegistry()

vcs_provider: Final[Callable[[str], Callable]] = vcs_providers.register


class DefaultVcsProvider(VcsProvider):
    """"""

    @property
    def is_available(self) -> bool:
        """"""
        return True

    @property
    def is_clean(self) -> bool:
        """"""
        return True

    def check_in_items(self, message: str, *files: tuple[Path, ...]) -> None:
        """

        Parameters
        ----------
        message: str :

        *files: tuple[Path, ...] :


        Returns
        -------

        """
        # Must not be provided
        pass

    def create_tag_from_string(self, version_formatted: str) -> None:
        """

        Parameters
        ----------
        version_formatted: str :


        Returns
        -------

        """
        # Must not be provided
        pass

    def get_most_recent_tag(self) -> Optional[Version]:
        """"""
        # Cannot be provided
        return None

    def get_number_of_changes_since_last_release(self) -> int:
        """"""
        # Cannot be provided
        return 0

    def get_changes_not_checked_in(self) -> int:
        """"""
        # Cannot be provided
        return 0

    def get_history(self, since_last_tag: bool = True) -> History:
        return History([])


class VcsProviderError(Exception):
    """"""

    pass



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""

from abc import ABC, abstractmethod
from enum import IntEnum
from functools import cached_property
from typing import Optional, Union, cast

from ..core.logging import logger
from ..core.version import Pep440VersionFormatter, Version
from ..vcs import CommitStatistics, CommitType
from .base import VersionModifier
from .increment import (
    DevelopmentVersionIncrementingVersionModifier,
    EpochIncrementingVersionModifier,
    MajorIncrementingVersionModifier,
    MicroIncrementingVersionModifier,
    MinorIncrementingVersionModifier,
    PostVersionIncrementingVersionModifier,
    ResetNonSemanticPartsModifier,
)
from .poetry_like import PoetryLikePreReleaseVersionModifier
from .preview import PreReleaseIncrementingVersionModifier


class _NoopVersionModifier(VersionModifier):
    def create_new_version(self) -> Version:
        return self.current_version


class Rating(IntEnum):
    """"""

    UNDEFINED = (0,)
    NOOP = (10,)
    LOCAL = (20,)
    DEVELOPMENT = 30
    POST = (40,)
    PRERELEASE = 50
    MICRO = 60
    MINOR = 70
    MAJOR = 80
    EPOCH = 90


class VersionPolicy(ABC):
    """"""

    def __init__(self, version: Version) -> None:
        """"""
        self.__version = version

    @abstractmethod
    def get_modifier(
        self, statistics: CommitStatistics, is_clean_repository: bool = True
    ) -> VersionModifier:
        """"""
        raise NotImplementedError()

    def save_version(self, version: Version) -> None:
        """"""
        version_formatted: str = Pep440VersionFormatter().format(version)
        logger.debug("Would save version %s", version_formatted)

    @property
    def _version(self) -> Version:
        return self.__version


class RatingBasedVersionPolicy(VersionPolicy):
    """"""

    def __init__(self, version: Version) -> None:
        """"""
        super().__init__(version)
        self.__version = version

    def get_modifier(
        self, statistics: CommitStatistics, is_clean_repository: bool = True
    ) -> VersionModifier:
        """"""
        if len(statistics.commit_type_count) == 0:
            return _NoopVersionModifier(self.__version, self)

        max_rating = self._get_max_rating(statistics, is_clean_repository)
        logger.debug(
            "Running with a version rating for the history of %s", max_rating
        )

        modifier: VersionModifier = _NoopVersionModifier(self.__version, self)

        if max_rating >= Rating.EPOCH:
            modifier = EpochIncrementingVersionModifier(self.__version, self)
        elif max_rating >= Rating.MAJOR:
            modifier = MajorIncrementingVersionModifier(self.__version, self)
        elif max_rating >= Rating.MINOR:
            modifier = self.__select_minor_version_modifier()
        elif max_rating >= Rating.MICRO:
            modifier = self.__select_micro_version_modifier()
        elif max_rating >= Rating.PRERELEASE:
            modifier = PoetryLikePreReleaseVersionModifier(
                self.__version, self
            )
        elif max_rating >= Rating.POST:
            modifier = PostVersionIncrementingVersionModifier(
                self.__version, self
            )
        elif max_rating >= Rating.DEVELOPMENT:
            modifier = DevelopmentVersionIncrementingVersionModifier(
                self.__version, self
            )
        elif max_rating >= Rating.LOCAL:
            raise NotImplementedError()

        logger.debug("Returning modifier %s", modifier)
        return modifier

    def _get_max_rating(self, statistics, is_clean_repository) -> int:
        """"""
        max_rating: int = -1
        current_rating: Union[Rating, int]
        for commit_type, count in statistics.commit_type_count.items():
            current_rating = self._rate_commit_type(commit_type)
            logger.debug(
                "Found %i commit(s) of type %s rated as %s",
                count,
                commit_type.name,
                current_rating,
            )

            if isinstance(current_rating, Rating):
                current_rating = current_rating.value
            current_rating = cast(int, current_rating)
            max_rating = max(max_rating, current_rating)

        logger.debug(
            "After evaluating all commits, the rating is set to %i", max_rating
        )

        if statistics.contains_breaking_changes:
            current_rating = self._rate_breaking_change()
            logger.debug(
                "Found at least one breaking change rated as %s",
                current_rating,
            )
            if isinstance(current_rating, Rating):
                current_rating = current_rating.value
            current_rating = cast(int, current_rating)
            max_rating = max(max_rating, current_rating)

        if not is_clean_repository:
            current_rating = self._rate_dirty_repository()
            logger.debug(
                "Running on a dirty repository rated as %s", current_rating
            )
            if isinstance(current_rating, Rating):
                current_rating = current_rating.value
            current_rating = cast(int, current_rating)
            max_rating = max(max_rating, current_rating)

        logger.debug("Rating set to %i", max_rating)
        return max_rating

    def __select_micro_version_modifier(self):
        modifier = MicroIncrementingVersionModifier(self.__version, self)
        # During development phases, micro-changes will
        # simply increment the pre-release
        if self.__version.is_pre_release:
            if not self.__version.is_development_version:
                pre_release_part = self.__select_preview_part(
                    self.__version.preview
                )
                modifier = PreReleaseIncrementingVersionModifier(
                    self.__version, self, pre_release_part, False
                )
            else:
                modifier = ResetNonSemanticPartsModifier(self.__version, self)
        return modifier

    def __select_minor_version_modifier(self):
        modifier = MinorIncrementingVersionModifier(self.__version, self)
        # During development, it might actually shift only the kind of pre-
        # release instead of increasing the minor version
        if self.__version.is_pre_release:
            if self.__version.is_development_version:
                modifier = ResetNonSemanticPartsModifier(self.__version, self)
            else:
                pre_release_part = self.__select_preview_part(
                    self.__version.preview
                )
                pre_release_part = self.__shift_preview_part(pre_release_part)
                modifier = PreReleaseIncrementingVersionModifier(
                    self.__version, self, pre_release_part, False
                )
        return modifier

    def __shift_preview_part(self, pre_release_part: str) -> str:
        if pre_release_part in ("a", "alpha"):
            return "beta"
        return pre_release_part

    def __select_preview_part(
        self, pre_release_part: Optional[tuple[str, int]]
    ) -> str:
        if pre_release_part is not None:
            return pre_release_part[0]
        return "alpha"

    @abstractmethod
    def _rate_commit_type(self, c_type: CommitType) -> Union[Rating, int]:
        """"""
        raise NotImplementedError()

    def _rate_breaking_change(self) -> Union[Rating, int]:
        """"""
        return Rating.MAJOR

    def _rate_dirty_repository(self) -> Union[Rating, int]:
        """"""
        return Rating.LOCAL


class SetBasedVersionPolicy(RatingBasedVersionPolicy):
    """"""

    @property
    @abstractmethod
    def epoch_increments(self) -> frozenset[CommitType]:
        """"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def major_increments(self) -> frozenset[CommitType]:
        """"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def minor_increments(self) -> frozenset[CommitType]:
        """"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def micro_increments(self) -> frozenset[CommitType]:
        """"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def post_increments(self) -> frozenset[CommitType]:
        """"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def dev_increments(self) -> frozenset[CommitType]:
        """"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def local_increments(self) -> frozenset[CommitType]:
        """"""
        raise NotImplementedError()

    def _rate_commit_type(self, c_type: CommitType) -> Union[Rating, int]:
        rating: Rating = Rating.NOOP

        if c_type in self.epoch_increments:
            rating = Rating.EPOCH
        elif c_type in self.major_increments:
            rating = Rating.MAJOR
        elif c_type in self.minor_increments:
            rating = Rating.MINOR
        elif c_type in self.micro_increments:
            rating = Rating.MICRO
        elif c_type in self.post_increments:
            rating = Rating.POST
        elif c_type in self.dev_increments:
            rating = Rating.DEVELOPMENT
        elif c_type in self.local_increments:
            rating = Rating.LOCAL

        return rating


class SemanticVersionPolicy(SetBasedVersionPolicy):
    """"""

    @cached_property
    def epoch_increments(self) -> frozenset[CommitType]:
        """"""
        return frozenset()

    @cached_property
    def major_increments(self) -> frozenset[CommitType]:
        """"""
        return frozenset()

    @cached_property
    def minor_increments(self) -> frozenset[CommitType]:
        """"""
        return frozenset(
            (
                CommitType.Feature,
                CommitType.Performance,
                CommitType.Refactoring,
            )
        )

    @cached_property
    def micro_increments(self) -> frozenset[CommitType]:
        """"""
        return frozenset(
            (CommitType.Bugfix, CommitType.Chore, CommitType.Documentation)
        )

    @cached_property
    def post_increments(self) -> frozenset[CommitType]:
        """"""
        return frozenset(
            (
                CommitType.Build,
                CommitType.CodeStyle,
                CommitType.ContinuousIntegration,
                CommitType.Test,
            )
        )

    @cached_property
    def dev_increments(self) -> frozenset[CommitType]:
        """"""
        return frozenset((CommitType.Undefined,))

    @cached_property
    def local_increments(self) -> frozenset[CommitType]:
        """"""
        return frozenset()



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""

from typing import Protocol, Union, cast, runtime_checkable

from .core.config import Config, ConfigKeys
from .core.logging import logger
from .core.version import Pep440VersionFormatter, Version


# Justification: Minimal protocol
class _ProjectWriter(Protocol):  # pylint: disable=R0903
    """"""

    def write(self, show_message: bool) -> None:
        """

        Parameters
        ----------
        show_message: bool :


        Returns
        -------

        """
        # Method empty: Only a protocol stub
        raise NotImplementedError()


@runtime_checkable
class _ProjectWriterClassic(Protocol):  # pylint: disable=R0903
    """"""

    def write_pyproject(self, show_message: bool) -> None:
        """

        Parameters
        ----------
        show_message: bool :


        Returns
        -------

        """
        # Method empty: Only a protocol stub
        raise NotImplementedError()


# Justification: Minimal protocol
@runtime_checkable
class _ProjectWriterHolder(Protocol):  # pylint: disable=R0903
    """"""

    pyproject: _ProjectWriter


# Justification: Implementation of minimal protocol
class StaticPep621VersionSource:  # pylint: disable=R0903
    """"""

    def __init__(
        self,
        _: Union[_ProjectWriterHolder, _ProjectWriterClassic],
        config: Config,
    ) -> None:
        self.__config = config

    @property
    def is_enabled(self) -> bool:
        """"""
        return (
            self.__config.get_pyproject_metadata(ConfigKeys.VERSION)
            is not None
        )

    def __get_current_version(self) -> Version:
        version: str = cast(
            str, self.__config.get_pyproject_metadata(ConfigKeys.VERSION)
        )
        return Version.from_string(version)

    def __set_current_version(self, ver: Version) -> None:
        formatter: Pep440VersionFormatter = Pep440VersionFormatter()
        version: str = formatter.format(ver)
        logger.debug("Setting new version %s", version)
        self.__config.set_pyproject_metadata(version, ConfigKeys.VERSION)

    current_version = property(__get_current_version, __set_current_version)
